import logging
import os
import time
from datetime import datetime
from logging import Logger
from typing import Literal
import sys
import numpy as np
import pandas as pd


class UnlabledInfoLoggingFormatter(logging.Formatter):
    """Logging formatter that omits level name for INFO messages."""

    def format(self, record):
        if record.levelname == "INFO":
            return record.getMessage()
        return f"[{record.levelname}] {record.getMessage()}"


def ensure_directory(dirpath: str) -> None:
    """Creates directory if not existing"""
    if not os.path.exists(dirpath):
        os.mkdir(dirpath)


def create_logger(logger_name: str, log_to_file: bool, debug: bool = False) -> Logger:
    """Creates logger with special handlers for console and optionally log files."""
    logger = logging.getLogger(logger_name)
    logger.propagate = False
    logger.handlers.clear()

    logger.setLevel(logging.DEBUG)

    # console logs
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(UnlabledInfoLoggingFormatter())
    if debug:
        console_handler.setLevel(logging.DEBUG)
    else:
        console_handler.setLevel(logging.INFO)
    logger.addHandler(console_handler)
    # file logs (optional)
    if log_to_file:
        ensure_directory("logs")

        log_filename = f"logs/ecdownload_{time.strftime('%Y%m%dT%H%M%S', time.localtime(time.time()))}.log"
        # Ensure that a new log is created instead of appending to an existing log
        new_log_filename = log_filename
        i = 2
        while os.path.exists(new_log_filename):
            new_log_filename = log_filename.replace(
                ".log",
                f"_{i}.log",
            )
            i = i + 1

        file_handler = logging.FileHandler(new_log_filename, mode="a")
        file_formatter = logging.Formatter(
            "[%(asctime)s] [%(levelname)s] %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
        )
        file_handler.setFormatter(file_formatter)
        file_handler.setLevel(logging.DEBUG)
        logger.addHandler(file_handler)
    return logger


def get_counter_message(
    counter: int | None = None, total_count: int | None = None
) -> tuple[str, int]:
    """Creates a formatted counter displaying current and total count (e.g. like this [ 7/10])."""
    max_count_digits = len(str(total_count))
    count_msg = ""
    if counter is not None and total_count is not None:
        count_msg += (
            "["
            + str(counter).rjust(max_count_digits)
            + "/"
            + str(total_count).rjust(max_count_digits)
            + "]"
        )
    elif counter is not None:
        count_msg += "[" + str(counter).rjust(max_count_digits) + "]"
    return count_msg, max_count_digits

def print_stdout(*values: object, sep: str = " ", end: str = "\n", flush: bool = True) -> None:
    """A print-like function using sys.stdout.write() to work with consolce and notebook outputs."""
    text = sep.join(str(v) for v in values) + end
    sys.stdout.write(text)
    if flush:
        sys.stdout.flush()

def console_exclusive_info(*values: object, end: str | None = "\n") -> None:
    """Wrapper for print function (forcibly flush the stream) and without logging"""
    print_stdout(*values, end=end, flush=True)


def log_textbox(
    text: str,
    logger: Logger,
    is_mayor: bool = False,
    line_length: int = 70,
    align: str | Literal["left", "center", "right"] | None = None,
    show_time: bool = False,
) -> None:
    if isinstance(align, str) and align not in ["left", "center", "right"]:
        raise ValueError(
            f'invalid value "{align}" for align, expected "left", "center" or "right"'
        )
    elif align is not None:
        raise TypeError(
            f"invalid type '{type(align).__name__}' for align, expected 'str' or None"
        )
    else:
        if is_mayor:
            align = "center"
        else:
            align = "left"

    top_left = "#" if is_mayor else "+"
    top_right = "#" if is_mayor else "+"
    bottom_right = "#" if is_mayor else "+"
    bottom_left = "#" if is_mayor else "+"
    vertical = "#" if is_mayor else "|"
    horizontal = "=" if is_mayor else "-"

    lines = text.split("\n")

    logger.info(top_left + horizontal * line_length + top_right)

    time_str: str = ""
    if show_time:
        time_str = pd.Timestamp.now().strftime("%Y-%m-%d %H:%M:%S") + " "

    for l in lines:
        if l == "---":
            logger.info(top_left + "-" * line_length + top_right)
            continue
        elif l == "===":
            logger.info(top_left + "-" * line_length + top_right)
            continue

        if align == "left":
            pad_left = 1
            pad_right = line_length - len(l) - 1
        elif align == "center":
            half_pad = (line_length - len(l)) / 2
            pad_left = int(np.floor(half_pad))
            pad_right = int(np.ceil(half_pad))
            # Only show time in left alignment
            time_str = ""
        else:
            pad_left = line_length - len(l) - 1
            pad_right = 1
            # Only show time in left alignment
            time_str = ""

        pad_right = max(0, pad_right - len(time_str))

        logger.info(
            vertical + " " * pad_left + l + " " * pad_right + time_str + vertical
        )
        # Only show time in first line
        time_str = ""

    logger.info(bottom_left + horizontal * line_length + bottom_right)



import numpy as np
from matplotlib.colors import Colormap, ListedColormap

from ..color import Color
from ..format_conversion import alpha_to_hex
from .cmap import Cmap


def get_cmap():
    colors = [
        "#c5c9c7",
        "#a2653e",
        "#ffffff",
        "#ff474c",
        "#0504aa",
        "#009337",
        "#840000",
        "#042e60",
        "#d8dcd6",
        "#ffff84",
        "#f5bf03",
        "#f97306",
        "#ff000d",
        "#5539cc",
        "#2976bb",
        "#0d75f8",
        "#014182",
        "#017b92",
        "#06b48b",
        "#aaff32",
        "#6dedfd",
        "#01f9c6",
        "#7bc8f6",
        "#d7fffe",
        "#a2cffe",
        "#04d9ff",
        "#7a9703",
        "#b2996e",
        "#ffbacd",
        "#d99b82",
        "#947e94",
        "#856798",
        "#ac86a8",
        "#59656d",
        "#76424e",
        "#363737",
    ]
    definitions = {
        -1: "unknown",
        0: "ground",
        1: "clear",
        2: "possible rain (clutter)",
        3: "possible snow (clutter)",
        4: "possible cloud (clutter)",
        5: "heavy rain",
        6: "heavy mixed-phase precipitation",
        7: "no rain or ice (possible liquid)",
        8: "liquid cloud",
        9: "drizzling liquid cloud",
        10: "warm rain",
        11: "cold rain",
        12: "melting snow",
        13: "snow (possible liquid)",
        14: "snow (no liquid)",
        15: "rimed snow (possible liquid)",
        16: "rimed snow and supercooled liquid",
        17: "snow and supercooled liquid",
        18: "supercooled liquid",
        19: "ice cloud (possible liquid)",
        20: "ice and supercooled liquid",
        21: "ice cloud (no liquid)",
        22: "stratospheric ice",
        23: "STS (PSC Type I)",
        24: "NAT (PSC Type II)",
        25: "insects",
        26: "dust",
        27: "sea salt",
        28: "continental pollution",
        29: "smoke",
        30: "dusty smoke",
        31: "dusty mix",
        32: "stratospheric ash",
        33: "stratospheric sulfate",
        34: "stratospheric smoke",
    }
    cmap = Cmap(colors=colors, name="synergistic_target_classification").to_categorical(
        definitions
    )
    return cmap



from typing import Iterable, Sequence

import numpy as np
import pandas as pd
import xarray as xr
from numpy.typing import NDArray

from ..constants import ALONG_TRACK_DIM, TIME_VAR
from ..np_array_utils import pad_true_sequence
from ..time import (
    TimeRangeLike,
    TimestampLike,
    time_to_iso,
    to_timestamp,
    validate_time_range,
)


def get_time_range(
    ds: xr.Dataset,
    time_range: TimeRangeLike | Iterable | None,
    time_var: str = TIME_VAR,
) -> tuple[pd.Timestamp, pd.Timestamp]:
    """
    Ensures a complete time range by filling in missing start or end values using dataset boundaries.

    Args:
        ds (xr.Dataset): Dataset containing the time variable.
        time_var (str): Name of the time variable in the dataset.
        time_range (TimeRangeLike | Iterable | None): A two-element list, tuple or array containing start and end times,
            which may be strings, pandas Timestamps, or None.

    Returns:
        List[pd.Timestamp]: A complete [start, end] time range as pandas Timestamps.
    """
    if isinstance(time_range, (Sequence, np.ndarray)) and not isinstance(
        time_range, str
    ):
        if len(time_range) >= 2:
            time_range = [
                time_range[0],
                time_range[-1],
            ]
        else:
            raise ValueError(f"invalid time range '{time_range}'")
    else:
        raise TypeError(
            f"Invalid type '{type(time_range).__name__}' for time_range. Expected a 2-element sequence (tuple or list)."
        )

    new_time_range: list[pd.Timestamp] = [pd.Timestamp(0), pd.Timestamp(0)]
    if time_range[0] is None:
        new_time_range[0] = to_timestamp(ds[time_var].values[0])
    else:
        new_time_range[0] = to_timestamp(time_range[0])

    if time_range[1] is None:
        new_time_range[1] = to_timestamp(ds[time_var].values[-1])
    else:
        new_time_range[1] = to_timestamp(time_range[1])

    return (new_time_range[0], new_time_range[1])


def filter_time(
    ds: xr.Dataset,
    time_range: TimeRangeLike | Iterable | None = None,
    timestamp: TimestampLike | None = None,
    only_center: bool = False,
    time_var: str = TIME_VAR,
    dim_var: str = ALONG_TRACK_DIM,
    pad_idxs: int = 0,
) -> xr.Dataset:
    """
    Filters an xarray Dataset to include only samples within a given time range.

    Args:
        ds (xr.Dataset): The input dataset containing a time coordinate.
        time_range (TimeRangeLike | Iterable | None):
            Start and end time of the range to filter, as strings or pandas timestamps.
        time_var (str, optional): Name of the time variable in `ds`. Defaults to TIME_VAR.
        dim_var (str, optional): Dimension name along which time is defined. Defaults to ALONG_TRACK_DIM.

    Returns:
        xr.Dataset: Subset of `ds` containing only samples within the specified time range.
    """
    if time_range is None and timestamp is None:
        raise ValueError(
            f"Can not use both arguments time_range and timestamp at the same time."
        )

    times = ds[time_var].values
    mask: NDArray[np.bool_] = np.full(times.shape, False, dtype=bool)
    if timestamp is not None:
        timestamp = to_timestamp(timestamp)

        tmin = to_timestamp(times[0])
        tmax = to_timestamp(times[-1])

        if not tmin <= timestamp <= tmax:
            raise ValueError(
                f"Timestamp {timestamp} lies outside of the dataset's time range ({tmin} -> {tmax})"
            )

        idx = np.argmin(np.abs(times - timestamp.to_numpy()))
        mask[idx] = True
    else:
        time_range = get_time_range(ds, time_range=time_range, time_var=time_var)

        mask = (times >= np.min([time_range[0], time_range[1]])) & (
            times <= np.max([time_range[0], time_range[1]])
        )

    if only_center:
        mask_true_idxs = np.where(mask)[0]
        if len(mask_true_idxs) > 0:
            print("hi")
            idx_center = mask_true_idxs[len(mask_true_idxs) // 2]
            print(f"{idx_center=}")
            mask[:] = False
            mask[idx_center] = True

    mask = pad_true_sequence(mask, pad_idxs)

    if np.sum(mask) == 0:
        msg = (
            f"No data falls into the given time range!\n"
            f"In the dataset time ranges from {times[0]} to {times[-1]}.\n"
        )
        raise ValueError(msg)

    da_mask: xr.DataArray = xr.DataArray(mask, dims=[dim_var], name=time_var)

    ds_new: xr.Dataset = ds.copy().where(da_mask, drop=True)
    ds_new.attrs = ds.attrs.copy()
    ds_new.encoding = ds.encoding.copy()

    return ds_new



from typing import Sequence

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.axes import Axes
from matplotlib.figure import Figure


def create_column_subfigures(
    ncols: int,
    single_figsize: tuple[float, float] = (3, 8),
    margin: float = 0.0,
) -> tuple[Figure, list[Axes]]:
    """
    Creates a figure with multiple subfigures arranged as columns in a single row, each containing one Axes.

    Args:
        ncols (int): Number of subfigures (columns) to create.
        single_figsize (tuple[float, float], optional): Size (width, height) of each individual subfigure.
            Defaults to (3, 8).

    Returns:
        tuple[Figure, list[Axes]]: The parent figure and a list of Axes objects, one for each subfigure.
    """
    fig: Figure = plt.figure(
        figsize=(single_figsize[0] * ncols + (ncols - 1) * margin, single_figsize[1])
    )
    figs: np.ndarray
    if ncols == 1:
        figs = np.array([fig])
    else:
        width_ratios = [single_figsize[0]]
        for c in range(ncols - 1):
            width_ratios.extend([margin, single_figsize[0]])

        figs = fig.subfigures(
            1,
            ncols + (ncols - 1),
            wspace=0.0,
            hspace=0.0,
            width_ratios=width_ratios,
        )
    axs: list[Axes] = [
        f.add_subplot([0, 0, 1, 1]) for i, f in enumerate(figs) if i % 2 == 0
    ]

    return fig, axs



_file_type_aliases = dict(
    # Level 1
    ATLNOM1B="ATL_NOM_1B",
    ANOM="ATL_NOM_1B",
    ATLDCC1B="ATL_DCC_1B",
    ADCC="ATL_DCC_1B",
    ATLCSC1B="ATL_CSC_1B",
    ACSC="ATL_CSC_1B",
    ATLFSC1B="ATL_FSC_1B",
    AFSC="ATL_FSC_1B",
    MSINOM1B="MSI_NOM_1B",
    MNOM="MSI_NOM_1B",
    MSIBBS1B="MSI_BBS_1B",
    MBBS="MSI_BBS_1B",
    MSISD11B="MSI_SD1_1B",
    MSD1="MSI_SD1_1B",
    MSISD21B="MSI_SD2_1B",
    MSD2="MSI_SD2_1B",
    MSIRGR1C="MSI_RGR_1C",
    MRGR="MSI_RGR_1C",
    BBRNOM1B="BBR_NOM_1B",
    BNOM="BBR_NOM_1B",
    BBRSNG1B="BBR_SNG_1B",
    BSNG="BBR_SNG_1B",
    BBRSOL1B="BBR_SOL_1B",
    BSOL="BBR_SOL_1B",
    BBRLIN1B="BBR_LIN_1B",
    BLIN="BBR_LIN_1B",
    CPRNOM1B="CPR_NOM_1B",  # JAXA product
    CNOM="CPR_NOM_1B",  # JAXA product
    # Level 2a
    ATLFM2A="ATL_FM__2A",
    AFM="ATL_FM__2A",
    ATLAER2A="ATL_AER_2A",
    AAER="ATL_AER_2A",
    ATLICE2A="ATL_ICE_2A",
    AICE="ATL_ICE_2A",
    ATLTC2A="ATL_TC__2A",
    ATC="ATL_TC__2A",
    ATLEBD2A="ATL_EBD_2A",
    AEBD="ATL_EBD_2A",
    ATLCTH2A="ATL_CTH_2A",
    ACTH="ATL_CTH_2A",
    ATLALD2A="ATL_ALD_2A",
    AALD="ATL_ALD_2A",
    MSICM2A="MSI_CM__2A",
    MCM="MSI_CM__2A",
    MSICOP2A="MSI_COP_2A",
    MCOP="MSI_COP_2A",
    MSIAOT2A="MSI_AOT_2A",
    MAOT="MSI_AOT_2A",
    CPRFMR2A="CPR_FMR_2A",
    CFMR="CPR_FMR_2A",
    CPRCD2A="CPR_CD__2A",
    CCD="CPR_CD__2A",
    CPRTC2A="CPR_TC__2A",
    CTC="CPR_TC__2A",
    CPRCLD2A="CPR_CLD_2A",
    CCLD="CPR_CLD_2A",
    CPRAPC2A="CPR_APC_2A",
    CAPC="CPR_APC_2A",
    ATLCLA2A="ATL_CLA_2A",  # JAXA product
    ACLA="ATL_CLA_2A",  # JAXA product
    MSICLP2A="MSI_CLP_2A",  # JAXA product
    MCLP="MSI_CLP_2A",  # JAXA product
    CPRECO2A="CPR_ECO_2A",  # JAXA product
    CECO="CPR_ECO_2A",  # JAXA product
    CPRCLP2A="CPR_CLP_2A",  # JAXA product
    CCLP="CPR_CLP_2A",  # JAXA product
    # Level 2b
    AMMO2B="AM__MO__2B",
    AMMO="AM__MO__2B",
    AMCTH2B="AM__CTH_2B",
    AMCTH="AM__CTH_2B",
    AMACD2B="AM__ACD_2B",
    AMACD="AM__ACD_2B",
    ACTC2B="AC__TC__2B",
    ACTC="AC__TC__2B",
    BMRAD2B="BM__RAD_2B",
    BMRAD="BM__RAD_2B",
    BMAFLX2B="BMA_FLX_2B",
    BMAFLX="BMA_FLX_2B",
    ACMCAP2B="ACM_CAP_2B",
    ACMCAP="ACM_CAP_2B",
    ACMCOM2B="ACM_COM_2B",
    ACMCOM="ACM_COM_2B",
    ACMRT2B="ACM_RT__2B",
    ACMRT="ACM_RT__2B",
    ALLDF2B="ALL_DF__2B",
    ACMBDF="ALL_DF__2B",
    ALLDF="ALL_DF__2B",
    ALL3D2B="ALL_3D__2B",
    ACMB3D="ALL_3D__2B",
    ALL3D="ALL_3D__2B",
    ACCLP2B="AC__CLP_2B",  # JAXA product
    ACCLP="AC__CLP_2B",  # JAXA product
    ACMCLP2B="ACM_CLP_2B",  # JAXA product
    ACMCLP="ACM_CLP_2B",  # JAXA product
    ALLRAD2B="ALL_RAD_2B",  # JAXA product
    ACMBRAD="ALL_RAD_2B",  # JAXA product
    ALLRAD="ALL_RAD_2B",  # JAXA product
    # Auxiliary data
    AUXMET1D="AUX_MET_1D",
    XMET="AUX_MET_1D",
    AUXJSG1D="AUX_JSG_1D",
    XJSG="AUX_JSG_1D",
    # Orbit data
    MPLORBSCT="MPL_ORBSCT",
    MPLORBS="MPL_ORBSCT",
    ORBSCT="MPL_ORBSCT",
    AUXORBPRE="AUX_ORBPRE",
    XORBP="AUX_ORBPRE",
    ORBPRE="AUX_ORBPRE",
    AUXORBRES="AUX_ORBRES",
    XORBR="AUX_ORBRES",
    ORBRES="AUX_ORBRES",
)


def _format_file_type_string(file_type: str) -> str:
    try:
        cleaned_file_type = file_type.replace("-", "").replace("_", "").upper()
        formatted_file_type = _file_type_aliases[cleaned_file_type]
    except KeyError as e:
        raise KeyError(f"'{file_type}' is not a valid FileType alias.")
    return formatted_file_type



import warnings

import numpy as np
from numpy.typing import ArrayLike, NDArray


def nan_mean(a: ArrayLike, axis: int | None = None) -> NDArray | float:
    """Compute the mean while ignoring NaNs."""
    with warnings.catch_warnings():  # ignore warings about all-nan values
        warnings.simplefilter("ignore", category=RuntimeWarning)
        a = np.asarray(a)
        return np.nanmean(a, axis=axis)


def nan_std(a: ArrayLike, axis: int | None = None) -> NDArray | float:
    """Compute the standard deviation while ignoring NaNs."""
    with warnings.catch_warnings():  # ignore warings about all-nan values
        warnings.simplefilter("ignore", category=RuntimeWarning)
        a = np.asarray(a)
        return np.nanstd(a, axis=axis)


def nan_min(a: ArrayLike, axis: int | None = None) -> NDArray | float:
    """Compute the minimum while ignoring NaNs."""
    with warnings.catch_warnings():  # ignore warings about all-nan values
        warnings.simplefilter("ignore", category=RuntimeWarning)
        a = np.asarray(a)
        if len(a) > 0:
            return np.nanmin(a, axis=axis)
        else:
            return np.nan


def nan_max(a: ArrayLike, axis: int | None = None) -> NDArray | float:
    """Compute the maximum while ignoring NaNs."""
    with warnings.catch_warnings():  # ignore warings about all-nan values
        warnings.simplefilter("ignore", category=RuntimeWarning)
        a = np.asarray(a)
        if len(a) > 0:
            return np.nanmax(a, axis=axis)
        else:
            return np.nan


def nan_sem(a: ArrayLike, axis: int | None = None) -> NDArray | float:
    """Compute the standard error of the mean while ignoring NaNs."""
    with warnings.catch_warnings():  # ignore warings about all-nan values
        warnings.simplefilter("ignore", category=RuntimeWarning)
        a = np.asarray(a)
        return np.nanstd(a, axis=axis) / np.sqrt(np.size(a, axis=0))


def nan_rmse(predictions: ArrayLike, targets: ArrayLike) -> float:
    """Root mean squared error (RMSE)"""
    predictions = np.asarray(predictions)
    targets = np.asarray(targets)
    with warnings.catch_warnings():  # ignore warings about all-nan values
        warnings.simplefilter("ignore", category=RuntimeWarning)
        result = np.sqrt(np.nanmean((targets - predictions) ** 2))
    return result


def nan_mae(predictions: ArrayLike, targets: ArrayLike) -> float:
    """Mean absolute error (MAE)"""
    predictions = np.asarray(predictions)
    targets = np.asarray(targets)
    with warnings.catch_warnings():  # ignore warings about all-nan values
        warnings.simplefilter("ignore", category=RuntimeWarning)
        result = np.nanmean(np.abs(targets - predictions))
    return result


def nan_mean_diff(predictions: ArrayLike, targets: ArrayLike) -> float:
    """Mean of element-wise differences (i.e., `mean(target - prediction)`)."""
    predictions = np.asarray(predictions)
    targets = np.asarray(targets)
    with warnings.catch_warnings():  # ignore warings about all-nan values
        warnings.simplefilter("ignore", category=RuntimeWarning)
        result = np.nanmean(targets - predictions)
    return result


def nan_diff_of_means(predictions: ArrayLike, targets: ArrayLike) -> float:
    """Difference between means of target and prediction (i.e., `mean(target) - mean(prediction)`)."""
    predictions = np.asarray(predictions)
    targets = np.asarray(targets)
    with warnings.catch_warnings():  # ignore warings about all-nan values
        warnings.simplefilter("ignore", category=RuntimeWarning)
        result = np.nanmean(targets) - np.nanmean(predictions)
    return result



import re
import textwrap

import numpy as np


def format_float(f: float | int) -> str:
    """
    Format a float or integer to a string with one decimal place.

    Raises `TypeError` for invalid input type.
    """
    if isinstance(f, (float, int)):
        return "{:.1f}".format(f)
    raise TypeError(
        f"Given value `f` hat wrong type '{type(f).__name__}', expecting 'float' or 'int'"
    )


def wrap_label(label: str, width: int = 40) -> str:
    wrapped_label = label
    match = re.match(r"([^\[]+)(\[[^\]]+\])?(.*)", label)
    if match:
        var_name = match.group(1).strip()
        units = match.group(2) or ""
        extra = match.group(3).strip()

        _width = width
        while len(var_name) % _width < _width / 2 and _width > 10:
            _width -= 1

        wrapped_var_name = textwrap.fill(var_name, width=_width)
        current = len(wrapped_var_name) % width

        if current + len(units) + len(extra) <= width:
            wrapped_label = f"{wrapped_var_name} {units} {extra}".strip()
        else:
            wrapped_label = f"{wrapped_var_name}\n{units} {extra}".strip()
    else:
        while len(label) % width < width / 2 and width > 10:
            width -= 1

        wrapped_label = textwrap.fill(label, width=width)
    return wrapped_label



from typing import Iterable, Sequence, TypeAlias

import numpy as np
import numpy.typing as npt

Number: TypeAlias = float | int
NumericPairLike: TypeAlias = (
    tuple[Number, Number] | list[Number] | Sequence[Number] | npt.NDArray[np.number]
)
NumericPairNoneLike: TypeAlias = (
    tuple[Number | None, Number | None]
    | list[Number | None]
    | Sequence[Number | None]
    | npt.NDArray[np.number]
)

ValueRangeLike: TypeAlias = NumericPairLike | NumericPairNoneLike
DistanceRangeLike: TypeAlias = NumericPairLike
LatLonCoordsLike: TypeAlias = NumericPairLike


def validate_numeric_range(
    input: ValueRangeLike,
    fallback: tuple[Number, Number] | None = None,
) -> tuple[float, float]:
    """Validates that the input is a pair with exactly 2 numeric elements that monotonically increasing.

    Args:
        input (ValueRangeLike): A sequence of 2 numbers.
        fallback (tuple[Number, Number], optional): Used to replace None values in `input`.

    Returns:
        A tuple of monotonically increasing two floats.

    Raises:
        `TypeError` or `ValueError` if validation fails.
    """
    _pair: tuple[Number, Number] = validate_numeric_pair(input, fallback)

    if _pair[0] > _pair[1]:
        raise TypeError("Both elements must be numeric ('int' or 'float')")

    return _pair


def validate_numeric_pair(
    input: NumericPairLike | NumericPairNoneLike,
    fallback: tuple[Number, Number] | None = None,
) -> tuple[float, float]:
    """Validates that the input is a pair with exactly 2 numeric elements.

    Args:
        input (NumericPairLike | NumericPairNoneLike): A sequence of 2 numbers.
        fallback (tuple[Number, Number], optional): Used to replace None values in `input`.

    Returns:
        A tuple of two floats.

    Raises:
        `TypeError` or `ValueError` if validation fails.
    """
    if isinstance(input, np.ndarray):
        if input.ndim != 1 or input.shape[0] != 2:
            raise ValueError(f"Expected 1D array of length 2, got shape {input.shape}")
        pair = input.tolist()
    else:
        pair = list(input)

    if len(pair) != 2:
        raise ValueError(f"Expected exactly 2 elements, got {len(pair)}")

    if fallback and not isinstance(pair[0], Number):
        pair[0] = fallback[0]

    if fallback and not isinstance(pair[1], Number):
        pair[1] = fallback[1]

    if not all(isinstance(x, Number) for x in pair):
        raise TypeError("Both elements must be numeric ('int' or 'float')")

    _pair: tuple[Number, Number] = (
        float(pair[0]),
        float(pair[1]),
    )

    return _pair



from logging import Logger

from ._exceptions import InvalidInputError


def parse_selected_index(
    selected_index: int | None, logger: Logger | None = None
) -> int | None:
    """Converts 1-indexed selected_index to 0-indexed and raises InvalidInputError if it is 0"""
    try:
        if selected_index is None:
            return None
        else:
            if selected_index >= 1:
                selected_index = selected_index - 1
            elif selected_index == 0:
                raise InvalidInputError(
                    "The indices in the found files list start at 1."
                )
            return selected_index
    except InvalidInputError as e:
        if logger:
            logger.exception(e)
        raise



from dataclasses import dataclass
from logging import Logger

from ._request import get_request_json
from ._types import Entrypoint


@dataclass
class EOCollection:
    name: str
    url_queryables: str | None
    url_items: str | None


def _get_collections_url(entrypoint: Entrypoint, title: str | None = None) -> str:
    """Returns the URL of the collections search request."""
    if not isinstance(title, str) or title == "":
        title = "earthcare"
    return f"{entrypoint.value}/collections/?title={title}&limit=100&startRecord=1"


def get_available_collections(
    entrypoint: Entrypoint,
    title: str | None = None,
    logger: Logger | None = None,
) -> list[EOCollection]:
    """Requests EO collections from the Server and returns available ones."""
    url = _get_collections_url(entrypoint=entrypoint, title=title)

    data = get_request_json(url=url, logger=logger)

    eo_collections = []
    for collection in data.get("collections", []):
        collection_id = collection.get("id")
        url_queryables: str | None = None
        url_items: str | None = None

        for link in collection.get("links", []):
            _rel = link.get("rel")
            if _rel == "http://www.opengis.net/def/rel/ogc/1.0/queryables":
                url_queryables = link.get("href", None)
            if _rel == "items":
                url_items = link.get("href", None)
                if isinstance(url_items, str):
                    url_items = url_items.split("items")[0] + "items"
            if isinstance(url_queryables, str) and isinstance(url_items, str):
                break

        eoc = EOCollection(
            name=collection_id,
            url_queryables=url_queryables,
            url_items=url_items,
        )
        eo_collections.append(eoc)

    return eo_collections



import logging
import math
import warnings
from numbers import Number
from typing import Iterable, Literal, Sequence

logger: logging.Logger = logging.getLogger(__name__)

import cartopy.crs as ccrs  # type: ignore
import cartopy.feature as cfeature  # type: ignore
import cartopy.io.img_tiles as cimgt  # type: ignore
import matplotlib as mpl
import matplotlib.patheffects as pe
import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
import pandas as pd
import seaborn as sns
import xarray as xr
from cartopy.crs import Projection
from cartopy.feature.nightshade import Nightshade  # type: ignore
from cartopy.mpl.feature_artist import FeatureArtist  # type: ignore
from cartopy.mpl.geoaxes import _ViewClippedPathPatch  # type: ignore
from cartopy.mpl.gridliner import Gridliner  # type: ignore
from matplotlib.axes import Axes
from matplotlib.colorbar import Colorbar
from matplotlib.colors import Colormap, LogNorm, Normalize
from matplotlib.figure import Figure, SubFigure
from matplotlib.image import AxesImage
from matplotlib.offsetbox import AnchoredOffsetbox, AnchoredText
from matplotlib.patches import Rectangle
from matplotlib.text import Text
from numpy.typing import ArrayLike, NDArray
from owslib.wms import WebMapService  # type: ignore

from ...utils import GroundSite, all_in, get_ground_site
from ...utils.constants import *
from ...utils.constants import FIGURE_MAP_HEIGHT, FIGURE_MAP_WIDTH
from ...utils.geo import get_coords, haversine
from ...utils.geo.bbox import compute_bbox, pad_bbox
from ...utils.np_array_utils import (
    circular_nanmean,
    isascending,
    ismonotonic,
    normalize,
    wrap_to_interval,
)
from ...utils.overpass import get_overpass_info
from ...utils.time import (
    TimedeltaLike,
    TimeRangeLike,
    TimestampLike,
    time_to_iso,
    to_timedelta,
    to_timestamp,
    to_timestamps,
    validate_time_range,
)
from ...utils.typing import ValueRangeLike, validate_numeric_pair
from ...utils.xarray_utils import filter_radius, filter_time
from ..color import Cmap, Color, ColorLike, get_cmap
from ..save import save_plot
from ..text import shade_around_text
from .annotation import (
    add_text,
    add_text_overpass_info,
    add_title_earthcare_frame,
    add_title_earthcare_time,
    format_var_label,
)
from .colorbar import add_colorbar, add_vertical_colorbar
from .defaults import get_default_cmap, get_default_norm, get_default_rolling_mean


def get_central_coords(proj: ccrs.Projection) -> tuple[float | None, float | None]:
    """Returns the central latitude and longitude of a Cartopy projection."""
    params = proj.proj4_params
    return params.get("lat_0", None), params.get("lon_0", None)


def add_gray_stock_img(
    ax: Axes,
    cmap: Cmap | str = "gray",
    alpha: float = 0.3,
    vmin: float = 0.0,
    vmax: float = 1.0,
) -> AxesImage:
    img = ax.stock_img()  # type:ignore

    def rgb2gray(rgb):
        return np.dot(rgb[..., :3], [0.2989, 0.5870, 0.1140])

    new_img = rgb2gray(img.get_array())
    new_img = normalize(new_img)
    img.set_visible(False)
    cmap_gray = get_cmap(cmap)
    newcmp = cmap_gray

    return ax.imshow(
        new_img,
        origin="lower",
        extent=img.get_extent(),
        transform=ax.projection,  # type:ignore
        cmap=newcmp,
        alpha=alpha,
        vmin=vmin,
        vmax=vmax,
    )


def get_osm_lod(a: ArrayLike, b: ArrayLike) -> int:
    lod = 2
    distance_km = haversine(a, b, units="km")
    if distance_km < 25:
        lod = 12
    elif distance_km < 50:
        lod = 11
    elif distance_km < 100:
        lod = 10
    elif distance_km < 200:
        lod = 9
    elif distance_km < 300:
        lod = 8
    elif distance_km < 500:
        lod = 7
    elif distance_km < 750:
        lod = 6
    elif distance_km < 1250:
        lod = 5
    elif distance_km < 2000:
        lod = 4
    elif distance_km < 5500:
        lod = 3

    logger.debug(f"distance_km={distance_km}, lod={lod}")

    return int(lod)


def get_arrow_style(linewidth):
    return f"simple,head_width={0.3+(linewidth*0.15)},head_length={0.3+(linewidth*0.3)},tail_width=0"


def set_view(
    ax: Axes,
    proj: Projection,
    lats: ArrayLike,
    lons: ArrayLike,
    pad: float = 0.05,
    pad_xmin: float | None = None,
    pad_xmax: float | None = None,
    pad_ymin: float | None = None,
    pad_ymax: float | None = None,
) -> Axes:
    lons = np.array(lons).flatten()
    lats = np.array(lats).flatten()

    if pad_xmin is None:
        pad_xmin = pad
    if pad_xmax is None:
        pad_xmax = pad
    if pad_ymin is None:
        pad_ymin = pad
    if pad_ymax is None:
        pad_ymax = pad

    xypoints = proj.transform_points(ccrs.PlateCarree(), lons, lats)
    x = xypoints[:, 0]
    y = xypoints[:, 1]

    xmin = x.min()
    xmax = x.max()
    xd = np.abs(xmax - xmin)
    ymin = y.min()
    ymax = y.max()
    yd = np.abs(ymax - ymin)
    ax.set_xlim(xmin - xd * pad_xmin, xmax + xd * pad_xmax)
    ax.set_ylim(ymin - yd * pad_ymin, ymax + yd * pad_ymax)

    return ax


def _validate_figsize(figsize: tuple[float, float]) -> tuple[float, float]:
    if not isinstance(figsize, Iterable):
        raise TypeError(
            f"invalid type of figsize '{type(figsize).__name__}'. expected tuple of 2 numbers."
        )
    else:
        if len(figsize) != 2:
            raise ValueError(
                f"invalid figsize '{figsize}'. expected tuple of 2 numbers."
            )
        elif not isinstance(figsize[0], Number) or not isinstance(figsize[1], Number):
            raise TypeError(
                f"invalid type of figsize '{type(figsize).__name__}[{type(figsize[0]).__name__}, {type(figsize[1]).__name__}]'. expected tuple of 2 numbers."
            )
        return (figsize[0], figsize[1])


def _ensure_figure_and_main_axis(
    ax: Axes | None, figsize: tuple[float, float] | None = None, dpi: int | None = None
) -> tuple[Figure, Axes]:
    fig: Figure
    if isinstance(ax, Axes):
        fig = ax.get_figure()  # type: ignore
        if not isinstance(fig, (Figure, SubFigure)):
            raise ValueError(f"Invalid Figure")
    else:
        fig = plt.figure(figsize=figsize, dpi=dpi)
        ax = fig.add_axes((0.0, 0.0, 1.0, 1.0))
    return fig, ax


def _validate_projection(
    projection: Literal["platecarree", "perspective", "orthographic"] | ccrs.Projection,
) -> tuple[type, float | None, float | None]:
    projection_type: type = ccrs.Projection
    central_latitude: float | None = None
    central_longitude: float | None = None
    if isinstance(projection, ccrs.Projection):
        projection_type = type(projection)
        central_coords = get_central_coords(projection)
        central_longitude = central_coords[0]
        central_latitude = central_coords[1]
    elif isinstance(projection, str):
        if projection.lower() == "platecarree":
            projection_type = ccrs.PlateCarree
        elif projection.lower() == "perspective":
            projection_type = ccrs.NearsidePerspective
        elif projection.lower() == "orthographic":
            projection_type = ccrs.Orthographic
        elif projection.lower() == "oblique_mercator":
            projection_type = ccrs.ObliqueMercator
        elif projection.lower() == "stereographic":
            projection_type = ccrs.Stereographic
        else:
            raise TypeError(
                f'Invalid projection: "{projection}"". Expected "platecarree", "perspective", "orthographic" or a instance of cartopy.crs.Projection'
            )

    return (projection_type, central_latitude, central_longitude)


def _validate_pad(pad: float | Iterable) -> list:
    if isinstance(pad, (float, int)):
        return [float(pad), float(pad), float(pad), float(pad)]
    elif isinstance(pad, Iterable):
        pad = list(pad)
        if len(pad) == 4:
            return [pad[0], pad[1], pad[2], pad[3]]
        else:
            raise ValueError(
                f"pad has too few elements ({len(pad)}). expected 4 elements: [pad_xmin, pad_xmax, pad_ymin, pad_ymax] or a single 'float' number"
            )
    raise TypeError(
        f"invalid type for pad '{type(pad).__name__}'. valid types are 'float' and 'list'."
    )


class MapFigure:
    """Figure object for displaying EarthCARE satellite track and/or imager swaths on a global map.

    This class sets up a georeferenced map canvas using a range of cartographic projections and visual styles.
    It serves as the basis for plotting 2D swath data (e.g., from MSI) or simple satellite tracks, optionally
    with info labels, backgrounds, and other styling options.

    Args:
        ax (Axes | None, optional): Existing matplotlib axes to plot on; if not provided, new axes will be created. Defaults to None.
        figsize (tuple[float, float], optional): Figure size in inches. Defaults to (FIGURE_MAP_WIDTH, FIGURE_MAP_HEIGHT).
        dpi (int | None, optional): Resolution of the figure in dots per inch. Defaults to None.
        title (str | None, optional): Title to display on the map. Defaults to None.
        style (str, optional): Base map style to use; options include "none", "stock_img", "gray", "osm", "satellite", "mtg", "msg". Defaults to "gray".
        projection (str | Projection, optional): Map projection to use; options include "platecarree", "perspective", "orthographic", or a custom `cartopy.crs.Projection`. Defaults to `ccrs.Orthographic()`.
        central_latitude (float | None, optional): Latitude at the center of the projection. Defaults to None.
        central_longitude (float | None, optional): Longitude at the center of the projection. Defaults to None.
        grid_color (ColorLike | None, optional): Color of grid lines. Defaults to None.
        border_color (ColorLike | None, optional): Color of border box around the map. Defaults to None.
        coastline_color (ColorLike | None, optional): Color of coastlines. Defaults to None.
        show_grid (bool, optional): Whether to show latitude/longitude grid lines. Defaults to True.
        show_top_labels (bool, optional): Whether to show tick labels on the top axis. Defaults to True.
        show_bottom_labels (bool, optional): Whether to show tick labels on the bottom axis. Defaults to True.
        show_right_labels (bool, optional): Whether to show tick labels on the right axis. Defaults to True.
        show_left_labels (bool, optional): Whether to show tick labels on the left axis. Defaults to True.
        show_text_time (bool, optional): Whether to display a datetime info text above the plot. Defaults to True.
        show_text_frame (bool, optional): Whether to display a EarthCARE frame info text above the plot. Defaults to True.
        show_text_overpass (bool, optional): Whether to display ground site overpass info in the plot. Defaults to True.
        show_night_shade (bool, optional): Whether to overlay the nighttime shading based on `timestamp`. Defaults to True.
        timestamp (TimestampLike | None, optional): Time reference used for nightshade overlay. Defaults to None.
        extent (Iterable | None, optional): Map extent given as [lon_min, lon_max, lat_min, lat_max]; overrides auto zoom. Defaults to None.
        lod (int, optional): Level of detail for coastlines and grid elements; higher values reduce complexity. Defaults to 2.
        coastlines_resolution (str, optional): Resolution of coastlines to display; options are "10m", "50m", or "110m". Defaults to "110m".
        azimuth (float, optional): Rotation of the `cartopy.crs.ObliqueMercator` projection, in degrees (if used). Defaults to 0.
        pad (float | list[float], optional): Padding applied when selecting a map extent. Defaults to 0.05.
        background_alpha (float, optional): Transparency level of the background map style. Defaults to 1.0.
    """

    def __init__(
        self,
        ax: Axes | None = None,
        figsize: tuple[float, float] = (FIGURE_MAP_WIDTH, FIGURE_MAP_HEIGHT),
        dpi: int | None = None,
        title: str | None = None,
        style: (
            str | Literal["none", "stock_img", "gray", "osm", "satellite", "mtg", "msg"]
        ) = "gray",
        projection: (
            Literal["platecarree", "perspective", "orthographic"] | ccrs.Projection
        ) = ccrs.Orthographic(),
        central_latitude: float | None = None,
        central_longitude: float | None = None,
        grid_color: ColorLike | None = None,
        border_color: ColorLike | None = None,
        coastline_color: ColorLike | None = None,
        show_grid: bool = True,
        show_top_labels: bool = True,
        show_bottom_labels: bool = True,
        show_right_labels: bool = True,
        show_left_labels: bool = True,
        show_text_time: bool = True,
        show_text_frame: bool = True,
        show_text_overpass: bool = True,
        show_night_shade: bool = True,
        timestamp: TimestampLike | None = None,
        extent: Iterable | None = None,
        lod: int = 2,
        coastlines_resolution: Literal["10m", "50m", "110m"] = "110m",
        azimuth: float = 0,
        pad: float | list[float] = 0.05,
        background_alpha: float = 1.0,
    ):
        self.figsize = _validate_figsize(figsize)
        self.fig, self.ax = _ensure_figure_and_main_axis(ax, figsize=figsize, dpi=dpi)

        self.dpi = dpi
        self.title = title
        self.style = style
        self.grid_color = Color.from_optional(grid_color)
        self.border_color = Color.from_optional(border_color)
        self.coastline_color = Color.from_optional(coastline_color)
        self.show_grid = show_grid
        self.show_top_labels = show_top_labels
        self.show_bottom_labels = show_bottom_labels
        self.show_right_labels = show_right_labels
        self.show_left_labels = show_left_labels
        self.show_text_time = show_text_time
        self.show_text_frame = show_text_frame
        self.show_text_overpass = show_text_overpass
        if timestamp is not None:
            timestamp = to_timestamp(timestamp)
        self.timestamp = timestamp
        self.extent: list | None = None
        if isinstance(extent, Iterable):
            self.extent = list(extent)
        self.central_latitude = central_latitude
        self.central_longitude = central_longitude
        self.lod = lod
        self.coastlines_resolution = coastlines_resolution
        self.azimuth = azimuth
        self.colorbar: Colorbar | None = None
        self.pad = _validate_pad(pad)
        self.background_alpha = background_alpha

        self.projection_type, clat, clon = _validate_projection(projection)
        if self.central_latitude is None:
            self.central_latitude = clat
        if self.central_longitude is None:
            self.central_longitude = clon

        self.grid_lines: Gridliner | None = None

        self.show_night_shade = show_night_shade

        self._init_axes()

    def set_view(
        self, lats: ArrayLike, lons: ArrayLike, pad: float | Iterable | None = None
    ) -> Axes:
        if isinstance(pad, (float | int | Iterable)):
            self.pad = _validate_pad(pad)
        self.ax = set_view(
            self.ax,
            self.projection,
            lats,
            lons,
            pad_xmin=self.pad[0],
            pad_xmax=self.pad[1],
            pad_ymin=self.pad[2],
            pad_ymax=self.pad[3],
        )
        return self.ax

    def set_extent(
        self, extent: list | None = None, pad: float | Iterable | None = None
    ) -> "MapFigure":
        if isinstance(extent, Iterable):
            self.extent = extent
            self.set_view(
                lons=np.array(self.extent[0:2]),
                lats=np.array(self.extent[2:4]),
                pad=pad,
            )
        return self

    def _init_axes(self) -> None:
        if self.projection_type == ccrs.PlateCarree:
            self.projection = self.projection_type(self.central_longitude)
        elif self.projection_type == ccrs.NearsidePerspective:
            self.projection = self.projection_type(
                central_longitude=self.central_longitude,
                central_latitude=self.central_latitude,
            )
        elif self.projection_type == ccrs.Orthographic:
            self.projection = self.projection_type(
                central_longitude=self.central_longitude,
                central_latitude=self.central_latitude,
            )
        elif self.projection_type == ccrs.ObliqueMercator:
            if self.central_longitude is None:
                self.central_longitude = 0
            if self.central_latitude is None:
                self.central_latitude = 0
            self.projection = self.projection_type(
                central_longitude=self.central_longitude,
                central_latitude=self.central_latitude,
                azimuth=self.azimuth,
            )
        elif self.projection_type == ccrs.Stereographic:
            self.projection = self.projection_type(
                central_longitude=self.central_longitude,
                central_latitude=self.central_latitude,
            )
        else:
            self.projection = self.projection_type()
        self.transform = ccrs.Geodetic()  # ccrs.PlateCarree()

        # making sure axis projection is setup correctly
        if not isinstance(self.ax, Axes):
            self.fig, self.ax = plt.subplots(
                subplot_kw={"projection": self.projection}, figsize=self.figsize
            )
        elif not (
            hasattr(self.ax, "projection")
            and type(self.ax.projection) == type(self.projection)
        ):
            tmp = self.ax.get_figure()
            if not isinstance(tmp, (Figure, SubFigure)):
                raise ValueError(f"Invalid Figure")
            self.fig = tmp  # type: ignore
            self.ax = self.ax

            pos = self.ax.get_position()
            self.ax.remove()
            self.ax = self.fig.add_subplot(pos, projection=self.projection)  # type: ignore

        # self.ax.set_facecolor("white")
        # self.ax.set_facecolor("none")

        if self.title:
            self.fig.suptitle(self.title)

        self.ax.axis("equal")

        # Earth image
        grid_color = Color("#000000")
        coastline_color = Color("#000000")

        if not isinstance(self.style, str):
            raise TypeError(
                f"style has wrong type '{type(self.style).__name__}'. Expected 'str'"
            )
        if self.style == "none":
            pass
        elif self.style == "stock_img":
            img = self.ax.stock_img()  # type: ignore
            grid_color = Color("#3f4d53")
            coastline_color = Color("#537585")
        elif self.style == "gray":
            img = add_gray_stock_img(self.ax)
            grid_color = Color("#6d6d6db3")
            coastline_color = Color("#C0C0C0")
        elif self.style == "osm":
            request = cimgt.OSM()
            img = self.ax.add_image(
                request,
                self.lod,
                interpolation="spline36",
                regrid_shape=2000,
            )  # type: ignore
            grid_color = Color("#6d6d6db3")
            coastline_color = Color("#C0C0C0")
        elif self.style == "satellite":
            request = cimgt.QuadtreeTiles()
            img = self.ax.add_image(
                request,
                self.lod,
                interpolation="spline36",
                regrid_shape=2000,
            )  # type: ignore
            grid_color = Color("#C0C0C099")
            coastline_color = Color("#C0C0C099")
        else:
            if not isinstance(self.timestamp, pd.Timestamp):
                msg = f"Missing timestamp for {self.style.upper()} data request for 'https://view.eumetsat.int' (timestamp={self.timestamp})"
                warnings.warn(msg)
            else:
                if self.style == "mtg":
                    if self.timestamp < to_timestamp("2024-09-23T02:00"):
                        self.style = "msg"
                        msg = (
                            f"Switching to MSG since MTG is only available from 2024-09-23 02:00 UTC onwards"
                            f"(timestamp given: {time_to_iso(self.timestamp, format='%Y-%m-%d %H:%M:%S')})"
                        )
                        warnings.warn(msg)
                img = add_gray_stock_img(self.ax)
                grid_color = Color("#3f4d53")
                coastline_color = Color("white").blend(0.5)  # Color("#3f4d53")

                date_str = (
                    pd.Timestamp(self.timestamp, tz="UTC")
                    .round("h")
                    .isoformat()
                    .replace("+00:00", "Z")
                )
                # Connect to NASA GIBS
                url = "https://view.eumetsat.int/geoserver/ows"

                wms = WebMapService(url)
                if self.style == "mtg":
                    layer = "mtg_fd:rgb_geocolour"  # "mtg_fd:ir105_hrfi" #"mumi:worldcloudmap_ir108" #"MODIS_Terra_SurfaceReflectance_Bands143"
                elif self.style == "msg":
                    layer = "msg_fes:rgb_naturalenhncd"
                elif self.style == "nasa":
                    wms = WebMapService(
                        "https://gibs.earthdata.nasa.gov/wms/epsg4326/best/wms.cgi?",
                        version="1.1.1",
                    )
                    layer = "MODIS_Terra_CorrectedReflectance_TrueColor"
                else:
                    layer = self.style
                    # raise NotImplementedError()
                wms_kwargs = {
                    "time": date_str,
                }

                self.ax.add_wms(wms, layer, wms_kwargs=wms_kwargs)  # type: ignore

                # Overlay white transparent layer
                if self.background_alpha < 1.0:
                    width, height = 1024, 512
                    white_overlay = np.ones((height, width, 4))
                    white_overlay[..., 3] = 1 - self.background_alpha

                    self.ax.imshow(
                        white_overlay,
                        origin="upper",
                        transform=ccrs.PlateCarree(),
                    )
        # else:
        #     raise ValueError(
        #         f'invalid style "{self.style}". Valid styles are: "gray", "osm", "satellite"'
        #     )

        # Grid lines
        _grid_color = self.grid_color
        if _grid_color is None:
            _grid_color = grid_color

        _border_color = self.border_color
        if _border_color is None:
            _border_color = _grid_color

        _coastline_color = self.coastline_color
        if _coastline_color is None:
            _coastline_color = coastline_color

        if self.show_grid:
            self.grid_lines = self.ax.gridlines(draw_labels=True, color=_grid_color, linewidth=0.5, linestyle="dashed")  # type: ignore
            self.grid_lines.top_labels = self.show_top_labels
            self.grid_lines.bottom_labels = self.show_bottom_labels
            self.grid_lines.right_labels = self.show_right_labels
            self.grid_lines.left_labels = self.show_left_labels
        # self.ax.coastlines(  # type: ignore
        #     color=coastlines_color, resolution=self.coastlines_resolution
        # )  # type: ignore
        self.ax.add_feature(cfeature.COASTLINE, edgecolor=_coastline_color)  # type: ignore
        # self.ax.add_feature(  # type: ignore
        #     cfeature.BORDERS,
        #     linewidth=0.5,
        #     linestyle="solid",
        #     edgecolor=_coastline_color,
        # )  # type: ignore
        self.ax.spines["geo"].set_edgecolor(_border_color)

        # Night shade
        if self.timestamp is not None:
            self.timestamp = to_timestamp(self.timestamp)
            if self.show_night_shade:
                night_shade_alpha = 0.15
                night_shade_color = Color("#000000")
                self.ax.add_feature(  # type: ignore
                    Nightshade(
                        self.timestamp,
                        alpha=night_shade_alpha,
                        color=night_shade_color,
                        linewidth=0,
                    )
                )  # type: ignore

    def plot_track(
        self,
        latitude: NDArray,
        longitude: NDArray,
        marker: str | None = None,
        markersize: float | int | None = None,
        linestyle: str | None = None,
        linewidth: float | int = 2,
        color: Color | ColorLike | None = None,
        alpha: float | None = 1.0,
        highlight_first: bool = True,
        highlight_first_color: Color | None = None,
        highlight_last: bool = True,
        highlight_last_color: Color | None = None,
        zorder: int | None = 4,
    ) -> "MapFigure":
        latitude = np.asarray(latitude)
        longitude = np.asarray(longitude)

        color = Color.from_optional(color)
        highlight_first_color = Color.from_optional(highlight_first_color)
        highlight_last_color = Color.from_optional(highlight_last_color)

        p = self.ax.plot(
            longitude,
            latitude,
            marker=marker,
            markersize=markersize,
            linestyle=linestyle,
            linewidth=linewidth,
            zorder=zorder,
            transform=self.transform,
            color=color,
            alpha=alpha,
            markeredgewidth=linewidth,
        )
        color = p[0].get_color()  # type: ignore
        if highlight_first_color is None:
            highlight_first_color = color
        if highlight_last_color is None:
            highlight_last_color = color

        if highlight_first:
            self.ax.plot(
                [longitude[0]],
                [latitude[0]],
                marker="o",
                markersize=markersize,
                linestyle="none",
                zorder=zorder if zorder is not None else 4,
                transform=self.transform,
                color=highlight_first_color,
                alpha=alpha,
            )

        if highlight_last:
            tmp_i = 0
            for i in range(len(longitude)):
                if (longitude[-1], latitude[-1]) != (
                    longitude[-2 - i],
                    latitude[-2 - i],
                ):
                    tmp_i = -2 - i
                    break
            arrow_style = get_arrow_style(linewidth)
            self.ax.annotate(
                "",
                xy=(longitude[-1], latitude[-1]),
                xytext=(longitude[tmp_i], latitude[tmp_i]),
                transform=self.transform,
                clip_on=True,
                annotation_clip=True,
                arrowprops=dict(
                    arrowstyle=arrow_style,
                    color=highlight_last_color,
                    lw=linewidth,
                    shrinkA=0,
                    shrinkB=0,
                    alpha=alpha,
                    connectionstyle="arc3,rad=0",
                    mutation_scale=10,
                ),
                zorder=zorder,
            )
        return self

    def plot_text(
        self,
        latitude: int | float,
        longitude: int | float,
        text: str,
        color: Color | ColorLike | None = "black",
        text_side: Literal["left", "right"] = "left",
        zorder: int | float = 8,
        padding: str = "  ",
    ) -> "MapFigure":
        if isinstance(text_side, str):
            if text_side == "left":
                horizontalalignment = "right"
                text = f"{text}{padding}"
            elif text_side == "right":
                horizontalalignment = "left"
                text = f"{padding}{text}"
            else:
                raise ValueError(
                    f'got invalid text_side "{text_side}". expected "left" or "right".'
                )
        else:
            raise TypeError(
                f"""invalid type '{type(text_side).__name__}' for text_side. expected type 'str': "left" or "right"."""
            )

        t = self.ax.text(
            longitude,
            latitude,
            text,
            color=color,
            verticalalignment="center",
            horizontalalignment=horizontalalignment,
            transform=self.transform,
            zorder=zorder,
            clip_on=True,
        )
        t = shade_around_text(t, alpha=0.5, linewidth=3, color="white")
        return self

    def plot_point(
        self,
        latitude: int | float,
        longitude: int | float,
        marker: str | None = "s",
        markersize: int | float = 5,
        color: Color | ColorLike | None = "black",
        zorder: int | float = 4,
        text: str | None = None,
        text_color: Color | ColorLike | None = "black",
        text_side: Literal["left", "right"] = "right",
        text_zorder: int | float = 8,
        text_padding: str = "  ",
    ) -> "MapFigure":
        self.ax.plot(
            [longitude],
            [latitude],
            marker=marker,
            markersize=markersize,
            linestyle="none",
            transform=self.transform,
            color=color,
            zorder=zorder,
        )
        if isinstance(text, str):
            self.plot_text(
                latitude=latitude,
                longitude=longitude,
                text=text,
                color=text_color,
                text_side=text_side,
                zorder=text_zorder,
                padding=text_padding,
            )
        return self

    def plot_radius(
        self,
        latitude: int | float,
        longitude: int | float,
        radius_km: int | float,
        color: Color | ColorLike | None = "black",
        face_color: Color | ColorLike | None = "#FFFFFF00",
        edge_color: Color | ColorLike | None = None,
        text_color: Color | ColorLike | None = None,
        text: str | None = None,
        text_side: Literal["left", "right"] = "right",
        marker: str | None = "s",
        zorder: int | float = 4,
        text_zorder: int | float = 8,
    ) -> "MapFigure":
        color = Color.from_optional(color)
        face_color = Color.from_optional(face_color)
        edge_color = Color.from_optional(edge_color) or color
        text_color = Color.from_optional(text_color) or color

        # Draw circle
        # TODO: workaround to avoid annoying warnings, need to change this later!
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore", message="Approximating coordinate system*"
            )
            self.ax.tissot(  # type: ignore
                rad_km=radius_km,
                lons=longitude,
                lats=latitude,
                n_samples=128,
                facecolor=face_color,
                edgecolor=edge_color,
                zorder=zorder,
            )  # type: ignore

        # Draw center point
        self.plot_point(
            longitude=longitude,
            latitude=latitude,
            marker=marker,
            markersize=5,
            color=color,
            zorder=zorder,
            text=text,
            text_color=text_color,
            text_side=text_side,
            text_zorder=text_zorder,
            text_padding="  ",
        )

        return self

    def _plot_overpass(
        self,
        lat_selection: NDArray,
        lon_selection: NDArray,
        lat_total: NDArray,
        lon_total: NDArray,
        site: GroundSite,
        radius_km: int | float,
        color_selection: Color | ColorLike | None = "ec:earthcare",
        linewidth_selection: float = 3,
        linestyle_selection: str | None = "solid",
        color_total: Color | ColorLike | None = "ec:blue",
        linewidth_total: float = 2.5,
        linestyle_total: str | None = "solid",
        site_text_side: Literal["left", "right"] = "right",
        timestamp: pd.Timestamp | None = None,
        view: Literal["global", "data", "overpass"] = "overpass",
        show_highlights: bool = True,
    ) -> "MapFigure":
        lat_selection = np.asarray(lat_selection)
        lon_selection = np.asarray(lon_selection)
        lat_total = np.asarray(lat_total)
        lon_total = np.asarray(lon_total)

        site_lat = site.latitude
        site_lon = site.longitude
        site_alt = site.altitude
        site_name = site.name

        self.central_latitude = site_lat
        self.central_longitude = site_lon

        if view == "overpass":
            self.lod = get_osm_lod(
                (lat_selection[0], lon_selection[0]),
                (lat_selection[-1], lon_selection[-1]),
            )
            self.coastlines_resolution = "10m"
        elif view == "data":
            self.lod = get_osm_lod(
                (lat_total[0], lon_total[0]), (lat_total[-1], lon_total[-1])
            )
            self.coastlines_resolution = "50m"
        else:
            self.lod = 2
            self.coastlines_resolution = "110m"

        if timestamp is not None:
            self.timestamp = to_timestamp(timestamp)

        pos = self.ax.get_position()
        self.fig.delaxes(self.ax)
        self.ax = self.fig.add_axes(pos)  # type: ignore
        self._init_axes()

        # TODO: workaround to avoid annoying warnings, need to change this later!
        warnings.filterwarnings("ignore", message="Approximating coordinate system*")
        self.plot_radius(
            latitude=site_lat,
            longitude=site_lon,
            radius_km=radius_km,
            text=site_name,
            text_side=site_text_side,
        )

        highlight_last = False if view == "overpass" else True
        self.plot_track(
            latitude=lat_total,
            longitude=lon_total,
            linewidth=linewidth_total,
            linestyle=linestyle_total,
            highlight_first=show_highlights and False,
            highlight_last=show_highlights and highlight_last,
            color=color_total,
        )
        highlight_first = True if view == "overpass" else False
        highlight_last = True if view == "overpass" else False
        self.plot_track(
            latitude=lat_selection,
            longitude=lon_selection,
            linewidth=linewidth_selection,
            linestyle=linestyle_selection,
            highlight_first=show_highlights and highlight_first,
            highlight_last=show_highlights and highlight_last,
            color=color_selection,
        )

        self.ax.axis("equal")
        # if view == "overpass":
        #     extent = compute_bbox(np.vstack((lat_selection, lon_selection)).T)
        # else:
        #     extent = compute_bbox(np.vstack((lat_total, lon_total)).T)

        if view == "global":
            self.ax.set_global()  # type: ignore
        elif view == "overpass":
            zoom_radius_meters = radius_km * 1e3 * 1.3
            if isinstance(self.projection, ccrs.PlateCarree):
                self.set_view(lats=lat_selection, lons=lon_selection)
            else:
                self.ax.set_xlim(-zoom_radius_meters, zoom_radius_meters)
                self.ax.set_ylim(-zoom_radius_meters, zoom_radius_meters)
        elif view == "data":
            self.set_view(lats=lat_total, lons=lon_total)

        return self

    def ecplot(
        self,
        ds: xr.Dataset,
        var: str | None = None,
        *,
        lat_var: str = TRACK_LAT_VAR,
        lon_var: str = TRACK_LON_VAR,
        swath_lat_var: str = SWATH_LAT_VAR,
        swath_lon_var: str = SWATH_LON_VAR,
        time_var: str = TIME_VAR,
        along_track_dim: str = ALONG_TRACK_DIM,
        across_track_dim: str = ACROSS_TRACK_DIM,
        site: str | GroundSite | None = None,
        radius_km: float = 100.0,
        time_range: TimeRangeLike | None = None,
        view: Literal["global", "data", "overpass"] = "global",
        zoom_tmin: TimestampLike | None = None,
        zoom_tmax: TimestampLike | None = None,
        color: ColorLike | None = "ec:earthcare",
        linewidth: float = 3,
        linestyle: str | None = "solid",
        color2: ColorLike | None = "ec:blue",
        linewidth2: float | None = None,
        linestyle2: str | None = None,
        cmap: str | Cmap | None = None,
        zoom_radius_km: float | None = None,
        extent: list[float] | None = None,
        central_latitude: float | None = None,
        central_longitude: float | None = None,
        value_range: ValueRangeLike | None = None,
        log_scale: bool | None = None,
        norm: Normalize | None = None,
        colorbar: bool = True,
        pad: float | list[float] | None = None,
        show_text_time: bool | None = None,
        show_text_frame: bool | None = None,
        show_text_overpass: bool | None = None,
        cb_orientation: str | Literal["vertical", "horizontal"] = "horizontal",
        cb_position: str | Literal["left", "right", "top", "bottom"] = "bottom",
        cb_alignment: str | Literal["left", "center", "right"] = "center",
        cb_buffer: float = 1.02,
        cb_width_ratio: float | str = "2%",
        cb_height_ratio: float | str = "100%",
        selection_max_time_margin: (
            TimedeltaLike | Sequence[TimedeltaLike] | None
        ) = None,
    ) -> "MapFigure":
        """
        Plot the EarthCARE satellite track on a map, optionally showing a 2D swath variable if `var` is provided.

        This method collects all required data from an EarthCARE `xarray.Dataset`.
        If `var` is given, the corresponding swath variable is plotted on the map using a
        color scale. Otherwise, the satellite ground track is plotted as a colored line.
        If `time_range` or `site` is given, the selected track section within the selected time range or in proximity to ground sites are highlighted.

        Args:
            ds (xr.Dataset): The EarthCARE dataset from which data will be plotted.
            var (str | None, optional): Name of a 2D swath variable to plot. If None, only the satellite ground track is shown. Defaults to None.
            lat_var (str, optional): Name of the latitude variable for the along-track data. Defaults to TRACK_LAT_VAR.
            lon_var (str, optional): Name of the longitude variable for the along-track data. Defaults to TRACK_LON_VAR.
            swath_lat_var (str, optional): Name of the latitude variable for the swath. Defaults to SWATH_LAT_VAR.
            swath_lon_var (str, optional): Name of the longitude variable for the swath. Defaults to SWATH_LON_VAR.
            time_var (str, optional): Name of the time variable. Defaults to TIME_VAR.
            along_track_dim (str, optional): Dimension name representing the along-track direction. Defaults to ALONG_TRACK_DIM.
            across_track_dim (str, optional): Dimension name representing the across-track direction. Defaults to ACROSS_TRACK_DIM.
            site (str | GroundSite | None, optional): Highlights data within `radius_km` of a ground site (given either as a `GroundSite` object or name string); ignored if not set. Defaults to None.
            radius_km (float, optional): Radius around the ground site to highlight data from; ignored if `site` not set. Defaults to 100.0.
            time_range (TimeRangeLike | None, optional): Time range to highlight as selection area; ignored if `site` is set. Defaults to None.
            view (Literal["global", "data", "overpass"], optional): Map extent mode: "global" for full world, "data" for tight bounds, or "overpass" to zoom around `site` or time range. Defaults to "global".
            zoom_tmin (TimestampLike | None, optional): Optional lower time bound used for zooming map around track. Defaults to None.
            zoom_tmax (TimestampLike | None, optional): Optional upper time bound used for zooming map around track. Defaults to None.
            color (ColorLike | None, optional): Color used for selected section of the track or entire track if no selection. Defaults to "ec:earthcare".
            linewidth (float, optional): Line width for selected track section. Defaults to 3.
            linestyle (str | None, optional): Line style for selected track section. Defaults to "solid".
            color2 (ColorLike | None, optional): Color used for unselected sections of the track. Defaults to "ec:blue".
            linewidth2 (float, optional): Line width for unselected sections. Defaults to None.
            linestyle2 (str | None, optional): Line style for unselected sections. Defaults to None.
            cmap (str | Cmap | None, optional): Colormap to use when plotting a swath variable. Defaults to None.
            zoom_radius_km (float | None, optional): If set, overrides map extent derived from `view` to use a fixed radius around the site or selection. Defaults to None.
            extent (list[float] | None, optional): Map extent in the form [lon_min, lon_max, lat_min, lat_max]. If given, overrides map extent derived from `view`. Defaults to None.
            central_latitude (float | None, optional): Central latitude used for the map projection. Defaults to None.
            central_longitude (float | None, optional): Central longitude used for the map projection. Defaults to None.
            value_range (ValueRangeLike | None, optional): Min and max range for the variable values; ignored if `var` is None. Defaults to None.
            log_scale (bool | None, optional): Whether to apply a logarithmic color scale to the variable. Defaults to None.
            norm (Normalize | None, optional): Matplotlib norm to use for color scaling. Defaults to None.
            colorbar (bool, optional): Whether to display a colorbar for the variable. Defaults to True.
            pad (float | list[float] | None, optional): Padding around the map extent; ignored if `extent` is given. Defaults to None.
            show_text_time (bool | None, optional): Whether to display the UTC time start and end of the selected track. Defaults to None.
            show_text_frame (bool | None, optional): Whether to display EarthCARE frame information. Defaults to None.
            show_text_overpass (bool | None, optional): Whether to display overpass site name and related info. Defaults to None.
            cb_orientation (Literal["vertical", "horizontal"], optional): Orientation of the colorbar. Defaults to "horizontal".
            cb_position (Literal["left", "right", "top", "bottom"], optional): Position of the colorbar in the plot. Defaults to "bottom".
            cb_alignment (Literal["left", "center", "right"], optional): Horizontal alignment of the colorbar relative to the plot. Defaults to "center".
            cb_buffer (float, optional): Distance between plot and colorbar in axes units. Defaults to 1.02.
            cb_width_ratio (float | str, optional): Width of the colorbar relative to the plot area. Defaults to "2.5%".
            cb_height_ratio (float | str, optional): Height of the colorbar relative to the plot area. Defaults to "100%".

        Returns:
            MapFigure: The figure object containing the map with track or swath.

        Example:
            ```python
            import earthcarekit as eck

            filepath = "path/to/mydata/ECA_EXAE_ATL_NOM_1B_20250606T132535Z_20250606T150730Z_05813D.h5"
            with eck.read_product(filepath) as ds:
                mf = eck.MapFigure()
                mf = mf.ecplot(ds)
            ```
        """
        if pad is not None:
            self.pad = _validate_pad(pad)
        if show_text_time is not None:
            self.show_text_time = show_text_time
        if show_text_frame is not None:
            self.show_text_frame = show_text_frame
        if show_text_overpass is not None:
            self.show_text_overpass = show_text_overpass

        _lat_var: str = lat_var
        _lon_var: str = lon_var

        _linewidth: float = linewidth
        _linewidth2: float
        if isinstance(linewidth2, (float, int)):
            _linewidth2 = float(linewidth2)
        else:
            _linewidth2 = linewidth * 0.7

        if isinstance(var, str):
            _linewidth = linewidth * 0.5
            linestyle = "dashed"
            _linewidth2 = linewidth * 0.2
            if all_in(
                (along_track_dim, across_track_dim), [str(d) for d in ds[var].dims]
            ):
                _lat_var = swath_lat_var
                _lon_var = swath_lon_var

        _site: GroundSite | None = None
        if isinstance(site, GroundSite):
            _site = site
        elif isinstance(site, str):
            _site = get_ground_site(site)

        coords_whole_flight = get_coords(ds, lat_var=lat_var, lon_var=lon_var)

        if time_range is not None:
            if zoom_tmin is None and time_range[0] is not None:
                zoom_tmin = to_timestamp(time_range[0])
            if zoom_tmax is None and time_range[1] is not None:
                zoom_tmax = to_timestamp(time_range[1])

        if zoom_tmin or zoom_tmax:
            ds_zoomed_in = filter_time(ds, time_range=[zoom_tmin, zoom_tmax])
            coords_zoomed_in = get_coords(
                ds_zoomed_in, lat_var=_lat_var, lon_var=_lon_var, flatten=True
            )
            coords_zoomed_in_track = get_coords(
                ds_zoomed_in, lat_var=lat_var, lon_var=lon_var
            )
        else:
            coords_zoomed_in = coords_whole_flight
            coords_zoomed_in_track = get_coords(ds, lat_var=lat_var, lon_var=lon_var)

        if isinstance(_site, GroundSite):
            ds_overpass = filter_radius(
                ds,
                radius_km=radius_km,
                site=_site,
                lat_var=lat_var,
                lon_var=lon_var,
                along_track_dim=along_track_dim,
            )
            info_overpass = get_overpass_info(
                ds_overpass,
                site_radius_km=radius_km,
                site=_site,
                time_var=time_var,
                lat_var=lat_var,
                lon_var=lon_var,
                along_track_dim=along_track_dim,
            )

            _coords_whole_flight = coords_whole_flight.copy()
            _selection_max_time_margin: tuple[pd.Timedelta, pd.Timedelta] | None = None
            if selection_max_time_margin is not None:
                if isinstance(selection_max_time_margin, str):
                    _selection_max_time_margin = (
                        to_timedelta(selection_max_time_margin),
                        to_timedelta(selection_max_time_margin),
                    )
                elif isinstance(selection_max_time_margin, (Sequence, np.ndarray)):
                    _selection_max_time_margin = (
                        to_timedelta(selection_max_time_margin[0]),
                        to_timedelta(selection_max_time_margin[1]),
                    )
                else:
                    raise ValueError(
                        f"invalid selection_max_time_margin: {selection_max_time_margin}"
                    )

                _ds = filter_time(
                    ds=ds,
                    time_range=(
                        to_timestamp(ds_overpass[time_var].values[0])
                        - _selection_max_time_margin[0],
                        to_timestamp(ds_overpass[time_var].values[1])
                        + _selection_max_time_margin[1],
                    ),
                    time_var=time_var,
                )
                _coords_whole_flight = get_coords(_ds, lat_var=lat_var, lon_var=lon_var)

            coords_overpass = get_coords(ds_overpass, lat_var=lat_var, lon_var=lon_var)
            _ = self._plot_overpass(
                lat_selection=coords_overpass[:, 0],
                lon_selection=coords_overpass[:, 1],
                lat_total=_coords_whole_flight[:, 0],
                lon_total=_coords_whole_flight[:, 1],
                site=_site,
                radius_km=radius_km,
                view=view,
                timestamp=info_overpass.closest_time,
                color_selection=color,
                linewidth_selection=_linewidth,
                linestyle_selection=linestyle,
                color_total=color2,
                linewidth_total=_linewidth2,
                linestyle_total=linestyle2,
                show_highlights=view == "overpass",
            )

            if isinstance(_selection_max_time_margin, tuple):
                self.plot_track(
                    latitude=coords_whole_flight[:, 0],
                    longitude=coords_whole_flight[:, 1],
                    color="white",
                    linestyle="solid",
                    linewidth=2,
                    highlight_first=False,
                    highlight_last=True,
                    zorder=3,
                )

            if view == "overpass":
                if self.show_text_overpass:
                    add_text_overpass_info(self.ax, info_overpass)
            if self.show_text_time:
                add_title_earthcare_time(
                    self.ax, tmin=info_overpass.start_time, tmax=info_overpass.end_time
                )
        else:

            if isinstance(central_latitude, (int, float)):
                self.central_latitude = central_latitude
            else:
                self.central_latitude = np.nanmean(coords_zoomed_in_track[:, 0])

            if isinstance(central_longitude, (int, float)):
                self.central_longitude = central_longitude
            else:
                if not ismonotonic(coords_whole_flight[:, 0]):
                    self.central_longitude = coords_whole_flight[-1, 1]
                else:
                    self.central_longitude = circular_nanmean(coords_whole_flight[:, 1])

            logger.debug(
                f"Set central coords to (lat={self.central_latitude}, lon={self.central_longitude})"
            )

            time = ds[time_var].values
            timestamp = time[len(time) // 2]
            self.timestamp = to_timestamp(timestamp)

            if view != "global":
                self.lod = get_osm_lod(coords_zoomed_in[0], coords_zoomed_in[-1])
                if extent is None:
                    extent = compute_bbox(coords_zoomed_in)
                    self.extent = extent

            pos = self.ax.get_position()
            self.fig.delaxes(self.ax)
            self.ax = self.fig.add_axes(pos)  # type: ignore
            self._init_axes()

            if time_range is not None:
                if view == "global":
                    _highlight_last = True
                else:
                    _highlight_last = False
                _ = self.plot_track(
                    latitude=coords_whole_flight[:, 0],
                    longitude=coords_whole_flight[:, 1],
                    linewidth=_linewidth2,
                    linestyle=linestyle,
                    highlight_first=False,
                    highlight_last=_highlight_last,
                    color=color2,
                )
                if view == "global":
                    _highlight_last = False
                else:
                    _highlight_last = True
                _ = self.plot_track(
                    latitude=coords_zoomed_in_track[:, 0],
                    longitude=coords_zoomed_in_track[:, 1],
                    linewidth=_linewidth,
                    linestyle=linestyle,
                    highlight_first=False,
                    highlight_last=_highlight_last,
                    color=color,
                )
            else:
                _ = self.plot_track(
                    latitude=coords_whole_flight[:, 0],
                    longitude=coords_whole_flight[:, 1],
                    linewidth=_linewidth,
                    linestyle=linestyle,
                    highlight_first=False,
                    highlight_last=True,
                    color=color,
                )

            self.ax.axis("equal")
            if view == "global":
                self.ax.set_global()  # type: ignore
            else:
                self.set_view(lats=coords_zoomed_in[:, 0], lons=coords_zoomed_in[:, 1])

            if self.show_text_time:
                add_title_earthcare_time(self.ax, ds=ds, tmin=zoom_tmin, tmax=zoom_tmax)

        if isinstance(var, str):
            if cmap is None:
                cmap = get_default_cmap(var, ds)
            if norm is None:
                norm = get_default_norm(var, ds)
            lats = ds[swath_lat_var].values
            lons = ds[swath_lon_var].values
            values = ds[var].values
            label = getattr(ds[var], "long_name", "")
            units = getattr(ds[var], "units", "")
            _ = self.plot_swath(
                lats,
                lons,
                values,
                cmap=cmap,
                label=label,
                units=units,
                value_range=value_range,
                log_scale=log_scale,
                norm=norm,
                colorbar=colorbar,
                cb_orientation=cb_orientation,
                cb_position=cb_position,
                cb_alignment=cb_alignment,
                cb_buffer=cb_buffer,
                cb_width_ratio=cb_width_ratio,
                cb_height_ratio=cb_height_ratio,
            )

        # if zoom_tmin or zoom_tmax:
        #     extent = compute_bbox(coords_zoomed_in)
        #     self.ax.set_extent(extent, crs=ccrs.PlateCarree())  # type: ignore
        if self.show_text_frame:
            add_title_earthcare_frame(self.ax, ds=ds)

        self.zoom(extent=extent, radius_km=zoom_radius_km)

        return self

    def plot_swath(
        self,
        lats: NDArray,
        lons: NDArray,
        values: NDArray,
        label: str = "",
        units: str = "",
        cmap: str | Cmap | None = None,
        value_range: ValueRangeLike | None = None,
        log_scale: bool | None = None,
        norm: Normalize | None = None,
        colorbar: bool = True,
        cb_orientation: str | Literal["vertical", "horizontal"] = "horizontal",
        cb_position: str | Literal["left", "right", "top", "bottom"] = "bottom",
        cb_alignment: str | Literal["left", "center", "right"] = "center",
        cb_buffer: float = 1.02,
        cb_width_ratio: float | str = "2.5%",
        cb_height_ratio: float | str = "100%",
        show_swath_border: bool = True,
    ) -> "MapFigure":
        cmap = get_cmap(cmap)

        if isinstance(value_range, Iterable):
            if len(value_range) != 2:
                raise ValueError(
                    f"invalid `value_range`: {value_range}, expecting (vmin, vmax)"
                )
        else:
            value_range = (None, None)

        if isinstance(cmap, Cmap) and cmap.categorical == True:
            norm = cmap.norm
        elif isinstance(norm, Normalize):
            if log_scale == True and not isinstance(norm, LogNorm):
                norm = LogNorm(norm.vmin, norm.vmax)
            elif log_scale == False and isinstance(norm, LogNorm):
                norm = Normalize(norm.vmin, norm.vmax)
            if value_range[0] is not None:
                norm.vmin = value_range[0]
            if value_range[1] is not None:
                norm.vmax = value_range[1]
        else:
            if log_scale == True:
                norm = LogNorm(value_range[0], value_range[1])
            else:
                norm = Normalize(value_range[0], value_range[1])

        assert isinstance(norm, Normalize)
        value_range = (norm.vmin, norm.vmax)

        if len(values.shape) == 3 and values.shape[2] == 3:
            mesh = self.ax.pcolormesh(
                lons.T,
                lats.T,
                values,
                shading="auto",
                transform=ccrs.PlateCarree(),
            )
        else:
            mesh = self.ax.pcolormesh(
                lons,
                lats,
                values,
                cmap=cmap,
                norm=norm,
                shading="auto",
                transform=ccrs.PlateCarree(),
            )
            if colorbar:
                self.colorbar = add_colorbar(
                    self.fig,
                    self.ax,
                    mesh,
                    orientation=cb_orientation,
                    position=cb_position,
                    alignment=cb_alignment,
                    buffer=cb_buffer,
                    width_ratio=cb_width_ratio,
                    height_ratio=cb_height_ratio,
                    label=format_var_label(label, units),
                    cmap=cmap,
                )
        if show_swath_border:
            edgecolor = Color("white").set_alpha(0.5)
            _ = self.plot_track(
                lats[:, 0],
                lons[:, 0],
                highlight_first=False,
                highlight_last=False,
                color=edgecolor,
                linewidth=1,
            )
            _ = self.plot_track(
                lats[:, -1],
                lons[:, -1],
                highlight_first=False,
                highlight_last=False,
                color=edgecolor,
                linewidth=1,
            )
            _ = self.plot_track(
                lats[0, :],
                lons[0, :],
                highlight_first=False,
                highlight_last=False,
                color=edgecolor,
                linewidth=1,
            )
            _ = self.plot_track(
                lats[-1, :],
                lons[-1, :],
                highlight_first=False,
                highlight_last=False,
                color=edgecolor,
                linewidth=1,
            )

        return self

    def zoom(
        self, extent: ArrayLike | None = None, radius_km: float | None = None
    ) -> "MapFigure":
        radius_meters: float = 0

        if extent is not None:
            extent = np.asarray(extent)
            if extent.shape[0] != 4:
                ValueError(
                    f"'extent' has wrong size ({extent.shape[0]}), expecting size of 4 (min_lon, max_lon, min_lat, max_lat)"
                )
            lon_extent_km = haversine([extent[2], extent[0]], [extent[2], extent[1]])
            lat_extent_km = haversine([extent[2], extent[0]], [extent[3], extent[0]])
            radius_meters = np.max([lon_extent_km, lat_extent_km]) * 1e3

        if isinstance(radius_km, (int, float)):
            radius_meters = radius_km * 1e3

        if isinstance(self.projection, ccrs.PlateCarree) and extent is not None:
            self.ax.set_extent(extent, crs=ccrs.PlateCarree())  # type: ignore
        elif (
            not isinstance(self.projection, ccrs.PlateCarree) and radius_km is not None
        ):
            self.ax.set_xlim(-radius_meters, radius_meters)
            self.ax.set_ylim(-radius_meters, radius_meters)

        return self

    def to_texture(
        self, remove_images: bool = True, remove_features: bool = True
    ) -> "MapFigure":
        # Remove anchored text and other artist text objects
        for artist in reversed(self.ax.artists):
            if isinstance(artist, (Text, AnchoredOffsetbox)):
                artist.remove()

        # Completely remove axis ticks and labels
        self.ax.axis("off")

        # Remove white frame around figure
        self.fig.subplots_adjust(left=0, right=1, bottom=0, top=1)

        # Remove ticks, tick labels, and gridlines
        self.ax.set_xticks([])
        self.ax.set_yticks([])
        self.ax.xaxis.set_ticklabels([])
        self.ax.yaxis.set_ticklabels([])
        self.ax.grid(False)

        # Remove outline box around map
        self.ax.spines["geo"].set_visible(False)

        # Make the map fill the whole figure
        self.ax.set_position((0.0, 0.0, 1.0, 1.0))

        if self.colorbar:
            self.colorbar.remove()

        if self.grid_lines:
            self.grid_lines.remove()

        if remove_images:
            for img in self.ax.get_images():
                img.remove()

        if remove_features:
            for c in self.ax.get_children():
                if isinstance(c, FeatureArtist):
                    c.remove()

        # for c in self.ax.get_children():
        #     if isinstance(c, _ViewClippedPathPatch):
        #         c.set_alpha(0)

        for c in self.fig.get_children():
            if isinstance(c, Rectangle):
                c.set_alpha(0)

        self.ax.set_facecolor("none")

        return self

    def show(self):
        self.fig.tight_layout()
        self.fig.show()

    def save(self, filename: str = "", filepath: str | None = None, **kwargs):
        save_plot(fig=self.fig, filename=filename, filepath=filepath, **kwargs)



from .config import read_config, set_config
from .ground_sites import GroundSite, get_ground_site
from .np_array_utils import ismonotonic, isndarray
from .profile_data.profile_data import ProfileData
from .read import *
from .rolling_mean import *
from .set import all_in
from .swath_data.swath_data import SwathData
from .xarray_utils import filter_radius, filter_time



import warnings
from typing import Literal

import numpy as np
from numpy.typing import NDArray


def rolling_mean_1d(x: NDArray, w: int, is_pad: bool = True) -> NDArray:
    if w > len(x):
        return np.full(len(x), np.nan)

    windows = np.lib.stride_tricks.sliding_window_view(x, w)

    with warnings.catch_warnings():  # ignore warings about all-nan values
        warnings.simplefilter("ignore", category=RuntimeWarning)
        result = np.nanmean(windows, axis=-1)

    if is_pad:
        left_pad = np.full(w // 2, np.nan)
        right_pad = np.full((w - 1) // 2, np.nan)
        return np.concatenate([left_pad, result, right_pad])
    else:
        return result


def rolling_mean_2d(
    x: NDArray,
    w: int,
    axis: Literal[0, 1] = 1,
    is_pad: bool = True,
    is_keep_full_nan_along_axis: bool = False,
    full_nan_axis: int | None = None,
) -> NDArray:
    result = np.apply_along_axis(rolling_mean_1d, axis, x, w=w, is_pad=is_pad)

    if is_keep_full_nan_along_axis:
        if full_nan_axis is None:
            full_nan_axis = (axis + 1) % 2

        nan_mask = np.isnan(x)
        mask_full_nan_along_axis = np.all(nan_mask, axis=full_nan_axis)

        if full_nan_axis == 1:
            result[mask_full_nan_along_axis, :] = np.nan
        elif full_nan_axis == 0:
            result[:, mask_full_nan_along_axis] = np.nan

    return result



from typing import Literal, Sequence

import matplotlib.colors
import numpy as np
from matplotlib.colors import Colormap


def shift_mpl_colormap(
    cmap: Colormap,
    start: float = 0.0,
    midpoint: float = 0.5,
    stop: float = 1.0,
    name: str = "shifted_cmap",
):
    """Create a colormap with its center point shifted to a specified value.

    This function is useful for data with asymmetric ranges (e.g., negative min and
    positive max) where you want the center of the colormap to align with a specific
    value like zero.

    Args:
        cmap (Colormap): Colormap to be modified
        start (float): Lower bound of the colormap range (value between 0 and `midpoint`). Defaults to 0.0.
        midpoint (float): New center point of the colormap (value between 0 and 1). Defaults to 0.5.
            For data ranging from vmin to vmax where you want the center at value v,
            set midpoint = 1 - vmax/(vmax + abs(vmin))
        stop (float): Upper bound of the colormap range (value between `midpoint` and 1). Defaults to 1.0.
        name (str): Name of the new colormap. Defaults to "shifted_cmap".

    Returns:
        matplotlib.colors.LinearSegmentedColormap: New colormap with shifted center
    """

    color_dict: dict[str, list] = {
        "red": [],
        "green": [],
        "blue": [],
        "alpha": [],
    }

    regular_indices = np.linspace(start, stop, 257)

    shifted_indices = np.hstack(
        [
            np.linspace(0.0, midpoint, 128, endpoint=False),
            np.linspace(midpoint, 1.0, 129, endpoint=True),
        ]
    )

    for reg_idx, shift_idx in zip(regular_indices, shifted_indices):
        r, g, b, a = cmap(reg_idx)
        color_dict["red"].append((shift_idx, r, r))
        color_dict["green"].append((shift_idx, g, g))
        color_dict["blue"].append((shift_idx, b, b))
        color_dict["alpha"].append((shift_idx, a, a))

    segmentdata: dict[
        Literal["red", "green", "blue", "alpha"], Sequence[tuple[float, ...]]
    ] = dict(
        red=color_dict["red"],
        green=color_dict["green"],
        blue=color_dict["blue"],
        alpha=color_dict["alpha"],
    )

    new_cmap = matplotlib.colors.LinearSegmentedColormap(name, segmentdata)
    return new_cmap



import math
from typing import SupportsFloat

import numpy as np
from numpy.typing import NDArray

from ..constants import SEMI_MAJOR_AXIS_METERS, SEMI_MINOR_AXIS_METERS


def geo_to_ecef(
    lat: SupportsFloat,
    lon: SupportsFloat,
    alt: SupportsFloat,
    target_radius: float = 1.0,
    perfect_sphere: bool = True,
    semi_major: float = SEMI_MAJOR_AXIS_METERS,
    semi_minor: float = SEMI_MINOR_AXIS_METERS,
) -> tuple[float, float, float]:
    """
    Converts geodetic coordinates (i.e. latitude, longitude and altitude above ellipsoid)
    to Earth-centered, Earth-fixed (ECEF) coordinates (i.e. x, y and z in cartesian coordinates).

    Args:
        lat (float): Latitude angle north (positive) and south (negative) of the equator in degrees.
        lon (float): Longitude angle east (positive) and west (negative) of the prime meridian in degrees.
        alt (float): Height above above the Earth ellipsoid in meters.
        target_radius (float, optional): Target mean radius of the Earth ellipsoid in the new cartesian coordinate system. Defaults to 1.
        semi_major (float, optional): Semi-major axis of the Earth ellipsoid in meters. Defaults to 6378137 (WGS 84).
        semi_minor (float, optional): Semi-minor axis of the Earth ellipsoid in meters. Defaults to 6356752.314245 (WGS 84).

    Returns:
        coords (tuple[float, float, float]): 3D coordinates in meters (ECEF: A right-handed cartesian coordinate system that has its origin at the Earth's center and is fixed with respect to the Earth's rotation).
            - x (float): Point along the axis passing through the equator at the prime meridian (i.e. latitude = 0, longitude = 0 degrees).
            - y (float): Point along the axis passing through the equator 90 degrees east of the Prime Meridian (i.e. latitude = 0, longitude = 90 degrees).
            - z (float): Point along the axis passing through the north pole (i.e. latitude = 90 degrees).
    """
    lat = float(lat)
    lon = float(lon)
    alt = float(alt)

    sin = lambda x: math.sin(x)
    cos = lambda x: math.cos(x)
    sqrt = lambda x: math.sqrt(x)

    lat = math.radians(lat)
    lon = math.radians(lon)

    if perfect_sphere:
        # Calculate ECEF coordinates
        f = 1 - (semi_major / semi_major)  # Flattening of the ellipsoid
        N = semi_major / sqrt(
            1 - (f * sin(lat)) ** 2
        )  # Prime vertical radius of curvature

        x = (N + alt) * cos(lat) * cos(lon)
        y = (N + alt) * cos(lat) * sin(lon)
        z = ((semi_major**2 / semi_major**2) * N + alt) * sin(lat)

        # # Alternative
        # e2 = 1 - (semi_minor**2 / semi_major**2) # Square of the first numerical eccentricity of the ellipsoid
        # N = semi_major / sqrt(1 - e2 * (sin(lat) ** 2)) # Prime vertical radius of curvature

        # Scale ECEF coordinates to target radius
        R = ((semi_major + semi_major) / 2) / target_radius
        x = -x / R
        y = -y / R
        z = z / R
    else:
        # Calculate ECEF coordinates
        f = 1 - (semi_minor / semi_major)  # Flattening of the ellipsoid
        N = semi_major / sqrt(
            1 - (f * sin(lat)) ** 2
        )  # Prime vertical radius of curvature

        x = (N + alt) * cos(lat) * cos(lon)
        y = (N + alt) * cos(lat) * sin(lon)
        z = ((semi_minor**2 / semi_major**2) * N + alt) * sin(lat)

        # # Alternative
        # e2 = 1 - (semi_minor**2 / semi_major**2) # Square of the first numerical eccentricity of the ellipsoid
        # N = semi_major / sqrt(1 - e2 * (sin(lat) ** 2)) # Prime vertical radius of curvature

        # Scale ECEF coordinates to target radius
        R = ((semi_major + semi_minor) / 2) / target_radius
        x = -x / R
        y = -y / R
        z = z / R

    return x, y, z


def sequence_geo_to_ecef(
    lats: NDArray | list[SupportsFloat],
    lons: NDArray | list[SupportsFloat],
    alts: NDArray | list[SupportsFloat] | SupportsFloat | None = None,
    target_radius: float = 1.0,
    perfect_sphere: bool = True,
    semi_major: float = SEMI_MAJOR_AXIS_METERS,
    semi_minor: float = SEMI_MINOR_AXIS_METERS,
) -> NDArray:
    lats = np.asarray(lats)
    lons = np.asarray(lons)
    if alts is None:
        alts = np.zeros(lats.shape)
    elif isinstance(alts, (float, int)):
        alts = np.repeat(float(alts), lats.shape)
    else:
        alts = np.asarray(alts)
    coords: NDArray = np.stack((lats, lons, alts)).T
    xyz = np.array(
        [
            list(
                geo_to_ecef(
                    c[0],
                    c[1],
                    c[2],
                    target_radius=target_radius,
                    perfect_sphere=perfect_sphere,
                    semi_major=semi_major,
                    semi_minor=semi_minor,
                )
            )
            for c in coords
        ]
    )
    return xyz



from typing import Literal

import numpy as np
from matplotlib import ticker
from matplotlib.axes import Axes
from matplotlib.cm import ScalarMappable
from matplotlib.colorbar import Colorbar
from matplotlib.colors import BoundaryNorm
from matplotlib.figure import Figure, SubFigure
from mpl_toolkits.axes_grid1.inset_locator import inset_axes  # type: ignore
from numpy.typing import ArrayLike

from ..color import Cmap


def add_vertical_colorbar(
    fig: Figure | SubFigure,
    ax: Axes,
    data: ScalarMappable,
    label: str | None = None,
    ticks: ArrayLike | None = None,
    tick_labels: ArrayLike | None = None,
    position: Literal["left", "right"] = "right",
    horz_buffer: float = 0.025,
    width_ratio: float | str = "1.25%",
    height_ratio: float | str = "100%",
    cmap: Cmap | None = None,
) -> Colorbar:
    """Creates a vertical colorbar that streches to the height of the plot and keeps a set horizontal padding and width."""
    valid_fig_types = (Figure, SubFigure)
    if not isinstance(fig, valid_fig_types):
        raise TypeError(
            f"{add_vertical_colorbar.__name__}() for `fig` expected type '{Figure.__name__}' or '{SubFigure.__name__}' but got '{type(fig).__name__}' instead"
        )
    if not isinstance(ax, Axes):
        raise TypeError(
            f"{add_vertical_colorbar.__name__}() for `ax` expected type '{Axes.__name__}' but got '{type(ax).__name__}' instead"
        )
    if not isinstance(data, ScalarMappable):
        raise TypeError(
            f"{add_vertical_colorbar.__name__}() for `data` expected type '{ScalarMappable.__name__}' but got '{type(data).__name__}' instead"
        )
    if not isinstance(position, str):
        raise TypeError(
            f"{add_vertical_colorbar.__name__}() for `position` expected type '{str.__name__}' but got '{type(position).__name__}' instead"
        )

    if position == "right":
        bbox_left = 1 + horz_buffer
    elif position == "left":
        bbox_left = 0 - horz_buffer
    else:
        raise ValueError(
            f"Invald colorbar position: '{position}'. Set it to either 'left' or 'right'."
        )

    cax = inset_axes(
        ax,
        width=width_ratio,
        height=height_ratio,
        loc="center left",
        bbox_to_anchor=(bbox_left, 0, 1, 1),
        bbox_transform=ax.transAxes,
        borderpad=0,
    )

    if isinstance(cmap, Cmap) and cmap.categorical:
        cbar_bounds = np.arange(cmap.N + 1)
        cbar_norm = BoundaryNorm(cbar_bounds, cmap.N)
        sm = ScalarMappable(cmap=cmap, norm=cbar_norm)
        sm.set_array([])
        data = sm
        ticks = cmap.ticks
        tick_labels = cmap.labels

    cb = fig.colorbar(data, cax=cax, label=label, ticks=ticks, spacing="proportional")

    if tick_labels is not None:
        tick_labels = [str(l) for l in np.asarray(tick_labels)]
        cb.set_ticklabels(tick_labels)
        if (
            isinstance(data, ScalarMappable)
            and isinstance(cmap, Cmap)
            and cmap.categorical
        ):
            cb.solids.set_edgecolor("face")  # type: ignore
            cb.ax.tick_params(which="minor", size=0)
    else:
        if isinstance(cb.locator, ticker.AutoLocator):
            tick_locator = ticker.MaxNLocator(nbins="auto", steps=[1, 2, 2.5, 3, 5, 10])
            cb.locator = tick_locator
        if hasattr(cb.formatter, "set_useMathText"):
            cb.formatter.set_useMathText(True)
        cb.ax.yaxis.set_offset_position("left")
        cb.update_ticks()
        if hasattr(cb.formatter, "set_powerlimits"):
            cb.formatter.set_powerlimits((-2, 5))

    return cb


def add_colorbar(
    fig: Figure | SubFigure,
    ax: Axes,
    data: ScalarMappable,
    label: str | None = None,
    ticks: ArrayLike | None = None,
    tick_labels: ArrayLike | None = None,
    orientation: str | Literal["vertical", "horizontal"] = "vertical",
    position: str | Literal["left", "right", "top", "bottom"] = "right",
    alignment: str | Literal["left", "center", "right"] = "center",
    buffer: float = 0.025,
    width_ratio: float | str = "1.25%",
    height_ratio: float | str = "100%",
    cmap: Cmap | None = None,
) -> Colorbar:
    """
    Creates a colorbar (vertical or horizontal) alongside an axis with fixed placement and size.

    Args:
        fig (Figure | SubFigure): The parent figure.
        ax (Axes): The axis to which the colorbar is attached.
        data (ScalarMappable): The data for the colorbar.
        label (str, optional): Label for the colorbar.
        ticks (ArrayLike, optional): Tick positions.
        tick_labels (ArrayLike, optional): Tick labels.
        orientation (str): 'vertical' or 'horizontal'.
        position (str): Position relative to the axis ('left', 'right', 'top', 'bottom').
        alignment (str): Alignment of the colorbar ('left', 'center', 'right').
        buffer (float): Padding between axis and colorbar.
        width_ratio (float | str): Width (for vertical) or height (for horizontal) of the colorbar.
        height_ratio (float | str): Height (for vertical) or width (for horizontal) of the colorbar.
        cmap (Cmap, optional): A custom Cmap with categorical info.
    """
    if not isinstance(fig, (Figure, SubFigure)):
        raise TypeError(
            f"{add_colorbar.__name__}() expected `fig` to be a Figure or SubFigure"
        )
    if not isinstance(ax, Axes):
        raise TypeError(f"{add_colorbar.__name__}() expected `ax` to be an Axes")
    if not isinstance(data, ScalarMappable):
        raise TypeError(
            f"{add_colorbar.__name__}() expected `data` to be a ScalarMappable"
        )

    if not isinstance(alignment, str):
        raise TypeError(
            f"""alignment has invalid type '{type(alignment).__name__}', expected 'str' ("left", "center", "right")"""
        )
    elif alignment not in ["left", "center", "right"]:
        raise ValueError(
            f"""invalid value "{alignment}" for aligment, valid values are: "left", "center", "right"."""
        )

    bbox_anchor: tuple[float, float, float, float]
    if orientation == "vertical":
        if position == "right":
            bbox_anchor = (1 + buffer, 0, 1, 1)
            loc = f"left {alignment}"
        elif position == "left":
            bbox_anchor = (0 - buffer, 0, 1, 1)
            loc = f"right {alignment}"
        else:
            raise ValueError(
                "For vertical colorbars, position must be 'left' or 'right'."
            )
        cax = inset_axes(
            ax,
            width=width_ratio,
            height=height_ratio,
            loc=loc,
            bbox_to_anchor=bbox_anchor,
            bbox_transform=ax.transAxes,
            borderpad=0,
        )
    elif orientation == "horizontal":
        if position == "bottom":
            bbox_anchor = (0, -buffer, 1, 1)
            loc = f"upper {alignment}"
        elif position == "top":
            bbox_anchor = (0, 1 + buffer, 1, 1)
            loc = f"lower {alignment}"
        else:
            raise ValueError(
                "For horizontal colorbars, position must be 'top' or 'bottom'."
            )
        cax = inset_axes(
            ax,
            width=height_ratio,
            height=width_ratio,
            loc=loc,
            bbox_to_anchor=bbox_anchor,
            bbox_transform=ax.transAxes,
            borderpad=0,
        )
    else:
        raise ValueError("Orientation must be either 'vertical' or 'horizontal'.")

    # Handle categorical colormap
    if isinstance(cmap, Cmap) and cmap.categorical:
        cbar_bounds = np.arange(cmap.N + 1)
        cbar_norm = BoundaryNorm(cbar_bounds, cmap.N)
        sm = ScalarMappable(cmap=cmap, norm=cbar_norm)
        sm.set_array([])
        data = sm
        ticks = cmap.ticks
        tick_labels = cmap.labels

    cb = fig.colorbar(
        data,
        cax=cax,
        orientation=orientation,
        label=label,
        ticks=ticks,
        spacing="proportional",
    )

    if tick_labels is not None:
        cb.set_ticklabels([str(l) for l in np.asarray(tick_labels)])
        if (
            isinstance(data, ScalarMappable)
            and isinstance(cmap, Cmap)
            and cmap.categorical
        ):
            cb.solids.set_edgecolor("face")  # type: ignore
            cb.ax.tick_params(which="minor", size=0)
    else:
        if hasattr(cb.formatter, "set_useMathText"):
            cb.formatter.set_useMathText(True)
        if orientation == "vertical":
            cb.ax.yaxis.set_offset_position("left")
        cb.update_ticks()
        if hasattr(cb.formatter, "set_powerlimits"):
            cb.formatter.set_powerlimits((-2, 5))

    return cb



import xarray as xr


def convert_scalar_var_to_str(ds: xr.Dataset, var: str) -> xr.Dataset:
    """Converts a given scalar variable inside a `xarray.Dataset` to string."""
    val = ds[var].item()
    if isinstance(val, bytes):
        val = val.decode("utf-8")
    else:
        val = str(val)
    ds[var] = xr.DataArray(val)
    return ds



import warnings
from typing import Any, Iterable, Literal, Sequence

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import xarray as xr
from matplotlib import font_manager
from matplotlib.axes import Axes
from matplotlib.collections import PathCollection, PolyCollection
from matplotlib.colorbar import Colorbar
from matplotlib.colors import Colormap, LogNorm, Normalize
from matplotlib.dates import date2num
from matplotlib.figure import Figure, SubFigure
from matplotlib.lines import Line2D
from matplotlib.offsetbox import AnchoredOffsetbox, AnchoredText
from matplotlib.text import Text
from numpy.typing import ArrayLike, NDArray

from ...utils.constants import (
    ALONG_TRACK_DIM,
    ELEVATION_VAR,
    FIGURE_HEIGHT_CURTAIN,
    FIGURE_HEIGHT_LINE,
    FIGURE_WIDTH_CURTAIN,
    FIGURE_WIDTH_LINE,
    HEIGHT_VAR,
    TEMP_CELSIUS_VAR,
    TIME_VAR,
    TRACK_LAT_VAR,
    TRACK_LON_VAR,
    TROPOPAUSE_VAR,
)
from ...utils.ground_sites import GroundSite, get_ground_site
from ...utils.overpass import get_overpass_info
from ...utils.profile_data import (
    ProfileData,
    ensure_along_track_2d,
    ensure_vertical_2d,
    validate_profile_data_dimensions,
)
from ...utils.time import (
    TimedeltaLike,
    TimeRangeLike,
    TimestampLike,
    to_timedelta,
    to_timestamp,
    to_timestamps,
    validate_time_range,
)
from ...utils.typing import DistanceRangeLike, ValueRangeLike
from ..color import Cmap, Color, get_cmap
from ..save import save_plot
from ._plot_1d_integer_flag import plot_1d_integer_flag
from .along_track import AlongTrackAxisStyle, format_along_track_axis
from .annotation import (
    add_text,
    add_text_product_info,
    add_title_earthcare_frame,
    format_var_label,
)
from .colorbar import add_vertical_colorbar
from .defaults import get_default_cmap, get_default_norm, get_default_rolling_mean
from .height_ticks import format_height_ticks
from .ticks import format_numeric_ticks
from .value_range import select_value_range


class LineFigure:
    def __init__(
        self,
        ax: Axes | None = None,
        figsize: tuple[float, float] = (FIGURE_WIDTH_LINE, FIGURE_HEIGHT_LINE),
        dpi: int | None = None,
        title: str | None = None,
        ax_style_top: AlongTrackAxisStyle | str = "geo",
        ax_style_bottom: AlongTrackAxisStyle | str = "time",
        num_ticks: int = 10,
        show_value_left: bool = True,
        show_value_right: bool = False,
        mode: str | Literal["line", "scatter", "area"] = "line",
        show_grid: bool = True,
        grid_color: str | None = Color("lightgray"),
        grid_which: Literal["major", "minor", "both"] = "major",
        grid_axis: Literal["both", "x", "y"] = "both",
        grid_linestyle: str = "dashed",
        grid_linewidth: float = 1,
    ):
        self.fig: Figure
        if isinstance(ax, Axes):
            tmp = ax.get_figure()
            if not isinstance(tmp, (Figure, SubFigure)):
                raise ValueError(f"Invalid Figure")
            self.fig = tmp  # type: ignore
            self.ax = ax
        else:
            self.fig = plt.figure(figsize=figsize, dpi=dpi)
            self.ax = self.fig.add_subplot()  # add_axes((0.0, 0.0, 1.0, 1.0))
        self.title = title
        if self.title:
            self.fig.suptitle(self.title, y=1.4)

        self.ax_top: Axes
        self.ax_right: Axes
        self.selection_time_range: tuple[pd.Timestamp, pd.Timestamp] | None = None
        self.ax_style_top: AlongTrackAxisStyle = AlongTrackAxisStyle.from_input(
            ax_style_top
        )
        self.ax_style_bottom: AlongTrackAxisStyle = AlongTrackAxisStyle.from_input(
            ax_style_bottom
        )

        self.info_text: AnchoredText | None = None
        self.info_text_loc: str = "upper right"
        self.num_ticks = num_ticks
        self.show_value_left: bool = show_value_left
        self.show_value_right: bool = show_value_right
        self.mode: str | Literal["line", "scatter", "area"] = mode

        self.show_grid = show_grid
        self.grid_color = Color.from_optional(grid_color)
        self.grid_which = grid_which
        self.grid_axis = grid_axis
        self.grid_linestyle = grid_linestyle
        self.grid_linewidth = grid_linewidth

        self.ax_right = self.ax.twinx()
        self.ax_right.set_ylim(self.ax.get_ylim())
        self.ax_right.set_yticks([])

        self.ax_top = self.ax.twiny()
        self.ax_top.set_xlim(self.ax.get_xlim())

    def _set_info_text_loc(self, info_text_loc: str | None) -> None:
        if isinstance(info_text_loc, str):
            self.info_text_loc = info_text_loc

    def _set_axes(
        self,
        tmin: np.datetime64,
        tmax: np.datetime64,
        vmin: float,
        vmax: float,
        time: NDArray,
        tmin_original: np.datetime64 | None = None,
        tmax_original: np.datetime64 | None = None,
        longitude: NDArray | None = None,
        latitude: NDArray | None = None,
        ax_style_top: AlongTrackAxisStyle | str | None = None,
        ax_style_bottom: AlongTrackAxisStyle | str | None = None,
    ) -> "LineFigure":
        if ax_style_top is not None:
            self.ax_style_top = AlongTrackAxisStyle.from_input(self.ax_style_top)
        if ax_style_bottom is not None:
            self.ax_style_bottom = AlongTrackAxisStyle.from_input(self.ax_style_bottom)
        if not isinstance(tmin_original, np.datetime64):
            tmin_original = tmin
        if not isinstance(tmax_original, np.datetime64):
            tmax_original = tmax

        self.ax.set_xlim((tmin, tmax))  # type: ignore
        self.ax.set_ylim((vmin, vmax))

        self.ax.grid(
            visible=self.show_grid,
            which=self.grid_which,
            axis=self.grid_axis,
            color=self.grid_color,
            linestyle=self.grid_linestyle,
            linewidth=self.grid_linewidth,
        )

        self.ax_right.set_ylim(self.ax.get_ylim())
        self.ax_top.set_xlim(self.ax.get_xlim())

        format_along_track_axis(
            self.ax,
            self.ax_style_bottom,
            time,
            tmin,
            tmax,
            tmin_original,
            tmax_original,
            longitude,
            latitude,
            num_ticks=self.num_ticks,
        )
        format_along_track_axis(
            self.ax_top,
            self.ax_style_top,
            time,
            tmin,
            tmax,
            tmin_original,
            tmax_original,
            longitude,
            latitude,
            num_ticks=self.num_ticks,
        )
        return self

    def plot(
        self,
        *,
        values: NDArray | None = None,
        time: NDArray | None = None,
        latitude: NDArray | None = None,
        longitude: NDArray | None = None,
        # Common args for wrappers
        mode: str | Literal["line", "scatter", "area"] | None = None,
        value_range: ValueRangeLike | None = None,
        log_scale: bool | None = None,
        norm: Normalize | None = None,
        time_range: TimeRangeLike | None = None,
        label: str | None = None,
        units: str | None = None,
        color: str | None = Color("ec:blue"),
        alpha: float = 1.0,
        linestyle: str | None = "solid",
        linewidth: float | int | None = 2.0,
        marker: str | None = "s",
        markersize: float | int | None = 2.0,
        selection_time_range: TimeRangeLike | None = None,
        selection_color: str | None = Color("ec:earthcare"),
        selection_linestyle: str | None = "dashed",
        selection_linewidth: float | int | None = 2.5,
        selection_highlight: bool = False,
        selection_highlight_inverted: bool = True,
        selection_highlight_color: str | None = Color("white"),
        selection_highlight_alpha: float = 0.5,
        selection_max_time_margin: (
            TimedeltaLike | Sequence[TimedeltaLike] | None
        ) = None,
        ax_style_top: AlongTrackAxisStyle | str | None = None,
        ax_style_bottom: AlongTrackAxisStyle | str | None = None,
        mark_profiles_at: Sequence[TimestampLike] | None = None,
        classes: Sequence[int] | dict[int, str] | None = None,
        classes_kwargs: dict[str, Any] = {},
        **kwargs,
    ) -> "LineFigure":
        # Parse colors
        color = Color.from_optional(color)
        selection_color = Color.from_optional(selection_color)
        selection_highlight_color = Color.from_optional(selection_highlight_color)

        if isinstance(mode, str):
            if mode in ["line", "scatter", "area"]:
                self.mode = mode
            else:
                raise ValueError(
                    f'invalid `mode` "{mode}", expected either "line", "scatter" or "area"'
                )

        if isinstance(value_range, Iterable):
            if len(value_range) != 2:
                raise ValueError(
                    f"invalid `value_range`: {value_range}, expecting (vmin, vmax)"
                )
        else:
            value_range = (None, None)

        if isinstance(norm, Normalize):
            if log_scale == True and not isinstance(norm, LogNorm):
                norm = LogNorm(norm.vmin, norm.vmax)
            elif log_scale == False and isinstance(norm, LogNorm):
                norm = Normalize(norm.vmin, norm.vmax)
            if value_range[0] is not None:
                norm.vmin = value_range[0]
            if value_range[1] is not None:
                norm.vmax = value_range[1]
        else:
            if log_scale == True:
                norm = LogNorm(value_range[0], value_range[1])
            else:
                norm = Normalize(value_range[0], value_range[1])
        value_range = (norm.vmin, norm.vmax)

        values = np.asarray(values)
        time = np.asarray(time)
        if latitude is not None:
            latitude = np.asarray(latitude)
        if longitude is not None:
            longitude = np.asarray(longitude)

        # Validate inputs
        if len(values.shape) != 1:
            raise ValueError(
                f"Values must be 1D, but has {len(values.shape)} dimensions (shape={values.shape})"
            )

        tmin_original = time[0]
        tmax_original = time[-1]
        vmin_original = values[0]
        vmax_original = values[-1]

        if selection_time_range is not None:
            self.selection_time_range = validate_time_range(selection_time_range)
            _selection_max_time_margin: tuple[pd.Timedelta, pd.Timedelta] | None = None
            if isinstance(selection_max_time_margin, (Sequence, np.ndarray)):
                _selection_max_time_margin = (
                    to_timedelta(selection_max_time_margin[0]),
                    to_timedelta(selection_max_time_margin[1]),
                )
            elif selection_max_time_margin is not None:
                _selection_max_time_margin = (
                    to_timedelta(selection_max_time_margin),
                    to_timedelta(selection_max_time_margin),
                )

            if _selection_max_time_margin is not None:
                time_range = [
                    np.max(
                        [
                            time[0],
                            (
                                self.selection_time_range[0]
                                - _selection_max_time_margin[0]
                            ).to_datetime64(),
                        ]
                    ),
                    np.min(
                        [
                            time[-1],
                            (
                                self.selection_time_range[1]
                                + _selection_max_time_margin[1]
                            ).to_datetime64(),
                        ]
                    ),
                ]

        if time_range is not None:
            if isinstance(time_range, Iterable) and len(time_range) == 2:
                for i in [0, -1]:
                    time_range = list(time_range)
                    if time_range[i] is None:
                        time_range[i] = time[i]
                    time_range = tuple(time_range)  # type: ignore
        else:
            time_range = (time[0], time[-1])

        _value_range: tuple[float, float] = select_value_range(
            data=values,
            value_range=value_range,
            pad_frac=0.0,
            use_min_max=True,
        )

        tmin = np.datetime64(time_range[0])
        tmax = np.datetime64(time_range[1])

        vmin: float = _value_range[0]
        vmax: float = _value_range[1]

        x: NDArray = time
        y: NDArray = values

        if classes is not None:
            _yaxis_position = classes_kwargs.get("yaxis_position", "left")
            _is_left = _yaxis_position == "left"
            _label = format_var_label(label, units, label_len=30)

            plot_1d_integer_flag(
                ax=self.ax if _is_left else self.ax_right,
                ax2=self.ax_right if _is_left else self.ax,
                data=y,
                x=x,
                classes=classes,
                ax_label=_label,
                **classes_kwargs,
            )
        else:
            line: list[Line2D] | PathCollection | PolyCollection
            if "line" in self.mode:
                line = self.ax.plot(
                    x,
                    y,
                    marker="none",
                    linewidth=linewidth,
                    linestyle=linestyle,
                    color=color,
                    alpha=alpha,
                )
            elif "scatter" in self.mode:
                line = self.ax.scatter(
                    x,
                    y,
                    marker=marker,
                    s=markersize,
                    color=color,
                    alpha=alpha,
                )  # , **plot_kwargs)
            elif "area" in self.mode:
                line = self.ax.fill_between(
                    x,
                    [0] * x.shape[0],
                    y,
                    zorder=0,
                    color=color,
                    alpha=alpha,
                )
            else:
                raise ValueError(f"invalid `mode` {self.mode}")

            format_numeric_ticks(
                ax=self.ax,
                axis="y",
                label=format_var_label(label, units),
                show_label=self.show_value_left,
                show_values=self.show_value_left,
            )
            format_numeric_ticks(
                ax=self.ax_right,
                axis="y",
                label=format_var_label(label, units),
                show_label=self.show_value_right,
                show_values=self.show_value_right,
            )

        if selection_time_range is not None:
            if selection_highlight:
                if selection_highlight_inverted:
                    self.ax.axvspan(
                        tmin,  # type: ignore
                        self.selection_time_range[0],  # type: ignore
                        color=selection_highlight_color,
                        alpha=selection_highlight_alpha,
                    )
                    self.ax.axvspan(
                        self.selection_time_range[1],  # type: ignore
                        tmax,  # type: ignore
                        color=selection_highlight_color,
                        alpha=selection_highlight_alpha,
                    )
                else:
                    self.ax.axvspan(
                        self.selection_time_range[0],  # type: ignore
                        self.selection_time_range[1],  # type: ignore
                        color=selection_highlight_color,
                        alpha=selection_highlight_alpha,
                    )

            for t in self.selection_time_range:  # type: ignore
                self.ax.axvline(
                    x=t,  # type: ignore
                    color=selection_color,
                    linestyle=selection_linestyle,
                    linewidth=selection_linewidth,
                )

        self._set_axes(
            tmin=tmin,
            tmax=tmax,
            vmin=vmin,
            vmax=vmax,
            time=time,
            tmin_original=tmin_original,
            tmax_original=tmax_original,
            latitude=latitude,
            longitude=longitude,
            ax_style_top=ax_style_top,
            ax_style_bottom=ax_style_bottom,
        )

        if mark_profiles_at is not None:
            for t in to_timestamps(mark_profiles_at):
                self.ax.axvline(t, color=selection_color, linestyle="solid", linewidth=selection_linewidth)  # type: ignore

        return self

    def ecplot(
        self,
        ds: xr.Dataset,
        var: str,
        *,
        time_var: str = TIME_VAR,
        lat_var: str = TRACK_LAT_VAR,
        lon_var: str = TRACK_LON_VAR,
        along_track_dim: str = ALONG_TRACK_DIM,
        values: NDArray | None = None,
        time: NDArray | None = None,
        latitude: NDArray | None = None,
        longitude: NDArray | None = None,
        site: str | GroundSite | None = None,
        radius_km: float = 100.0,
        mark_closest_profile: bool = False,
        show_info: bool = True,
        info_text_loc: str | None = None,
        # Common args for wrappers
        mode: str | Literal["line", "scatter", "area"] | None = None,
        value_range: ValueRangeLike | None = None,
        log_scale: bool | None = None,
        norm: Normalize | None = None,
        time_range: TimeRangeLike | None = None,
        label: str | None = None,
        units: str | None = None,
        color: str | None = Color("ec:blue"),
        alpha: float = 1.0,
        linestyle: str | None = "solid",
        linewidth: float | int | None = 2.0,
        marker: str | None = "s",
        markersize: float | int | None = 2.0,
        selection_time_range: TimeRangeLike | None = None,
        selection_color: str | None = Color("ec:earthcare"),
        selection_linestyle: str | None = "dashed",
        selection_linewidth: float | int | None = 2.5,
        selection_highlight: bool = False,
        selection_highlight_inverted: bool = True,
        selection_highlight_color: str | None = Color("white"),
        selection_highlight_alpha: float = 0.5,
        selection_max_time_margin: (
            TimedeltaLike | Sequence[TimedeltaLike] | None
        ) = None,
        ax_style_top: AlongTrackAxisStyle | str | None = None,
        ax_style_bottom: AlongTrackAxisStyle | str | None = None,
        mark_profiles_at: Sequence[TimestampLike] | None = None,
        classes: Sequence[int] | dict[int, str] | None = None,
        classes_kwargs: dict[str, Any] = {},
        **kwargs,
    ) -> "LineFigure":
        # Collect all common args for wrapped plot function call
        local_args = locals()
        # Delete all args specific to this wrapper function
        del local_args["self"]
        del local_args["ds"]
        del local_args["var"]
        del local_args["time_var"]
        del local_args["lat_var"]
        del local_args["lon_var"]
        del local_args["along_track_dim"]
        del local_args["site"]
        del local_args["radius_km"]
        del local_args["show_info"]
        del local_args["info_text_loc"]
        del local_args["mark_closest_profile"]
        # Delete kwargs to then merge it with the residual common args
        del local_args["kwargs"]
        all_args = {**local_args, **kwargs}

        if all_args["values"] is None:
            all_args["values"] = ds[var].values
        if all_args["time"] is None:
            all_args["time"] = ds[time_var].values
        if all_args["latitude"] is None:
            all_args["latitude"] = ds[lat_var].values
        if all_args["longitude"] is None:
            all_args["longitude"] = ds[lon_var].values

        # Set default values depending on variable name
        if label is None:
            all_args["label"] = (
                "Values" if not hasattr(ds[var], "long_name") else ds[var].long_name
            )
        if units is None:
            all_args["units"] = "-" if not hasattr(ds[var], "units") else ds[var].units
        if classes is not None and len(classes) > 0:
            all_args["value_range"] = (-0.5, len(classes) - 0.5)
        elif value_range is None and log_scale is None and norm is None:
            all_args["norm"] = get_default_norm(var)

        # Handle overpass
        _site: GroundSite | None = None
        if isinstance(site, GroundSite):
            _site = site
        elif isinstance(site, str):
            _site = get_ground_site(site)
        else:
            pass

        if isinstance(_site, GroundSite):
            info_overpass = get_overpass_info(
                ds,
                site_radius_km=radius_km,
                site=_site,
                time_var=time_var,
                lat_var=lat_var,
                lon_var=lon_var,
                along_track_dim=along_track_dim,
            )
            overpass_time_range = info_overpass.time_range
            all_args["selection_time_range"] = overpass_time_range
            if mark_closest_profile:
                _mark_profiles_at = all_args["mark_profiles_at"]
                if isinstance(_mark_profiles_at, (Sequence, np.ndarray)):
                    list(_mark_profiles_at).append(info_overpass.closest_time)
                    all_args["mark_profiles_at"] = _mark_profiles_at
                else:
                    all_args["mark_profiles_at"] = [info_overpass.closest_time]

        self.plot(**all_args)

        self._set_info_text_loc(info_text_loc)
        if show_info:
            self.info_text = add_text_product_info(
                self.ax, ds, append_to=self.info_text, loc=self.info_text_loc
            )

        return self

    def to_texture(self) -> "LineFigure":
        # Remove anchored text and other artist text objects
        for artist in reversed(self.ax.artists):
            if isinstance(artist, (Text, AnchoredOffsetbox)):
                artist.remove()

        # Completely remove axis ticks and labels
        self.ax.axis("off")

        if self.ax_top:
            self.ax_top.axis("off")

        if self.ax_right:
            self.ax_right.axis("off")

        # Remove white frame around figure
        self.fig.subplots_adjust(left=0, right=1, bottom=0, top=1)

        return self

    def show(self):
        self.fig.tight_layout()
        self.fig.show()

    def save(self, filename: str = "", filepath: str | None = None, **kwargs):
        save_plot(fig=self.fig, filename=filename, filepath=filepath, **kwargs)



import warnings
from typing import Callable, Sequence

import numpy as np
import pandas as pd
import xarray as xr
from xarray import Dataset

from ...constants import ALONG_TRACK_DIM, TRACK_LON_VAR
from ...xarray_utils import concat_datasets
from ._generic import read_product
from .file_info import ProductDataFrame


def circular_mean_np(
    data: np.ndarray, axis: int = -1, degrees: bool = True
) -> np.ndarray:
    if degrees:
        data = np.deg2rad(data)
    sin_mean = np.mean(np.sin(data), axis=axis)
    cos_mean = np.mean(np.cos(data), axis=axis)
    angle = np.arctan2(sin_mean, cos_mean)
    return np.rad2deg(angle) if degrees else angle


def read_products(
    filepaths: Sequence[str] | pd.DataFrame,
    zoom_at: float | None = None,
    along_track_dim: str = ALONG_TRACK_DIM,
    func: Callable | None = None,
    func_inputs: Sequence[dict] | None = None,
    max_num_files: int = 8,
    coarsen: bool = True,
) -> Dataset:
    """Read and concatenate a sequence of EarthCARE frames into a single xarray Dataset.

    Optionally applies a processing function to each frame and zooms in on a specific region
    (defined by `zoom_at`). Handles coarse averaging when not zooming.

    Args:
        filepaths (Sequence[str] or pandas.DataFrame): EarthCARE product file paths as a list or a DataFrame
            with metadata including `filepath`, `orbit_number`, and `frame_id`.
        zoom_at (float, optional): If set, selects only a zoomed-in portion of the frames
            around this fractional index. Defaults to None.
        along_track_dim (str): Name of the dimension to concatenate along. Defaults to ALONG_TRACK_DIM.
        func (Callable, optional): Function to apply to each frame after loading. Defaults to None.
        func_inputs (Sequence[dict], optional): Optional per-frame arguments to pass to `func`. Defaults to None.
        max_num_files (int, optional): Max. number of files that are allowed to be loaded at once. A `ValueError` is raised if above. Defaults to 8 (e.g., full orbit).
        coarsen (bool): If Ture, read data sets are coarened depending on the number given of files. Only aplicable when not zooming. Defaults to Ture.

    Returns:
        Dataset: Concatenated dataset with all frames along `along_track_dim`.
    """
    if isinstance(filepaths, str):
        filepaths = [filepaths]
    elif isinstance(filepaths, pd.DataFrame):
        df = filepaths.sort_values(by="filepath")
        filepaths = df["filepath"].tolist()
    else:
        df = ProductDataFrame.from_files(list(filepaths)).sort_values(by="filepath")
        filepaths = df["filepath"].tolist()

    if len(filepaths) == 0:
        raise ValueError(f"Given sequence of product files paths is empty")
    elif len(filepaths) == 1:
        warnings.warn(f"Can not concatenate frames since only one file path was given")
        return read_product(filepaths[0])
    elif len(filepaths) > max_num_files:
        raise ValueError(
            f"Too many files provided: {len(filepaths)} (currently maximum allowed is {max_num_files}). "
            "Please reduce the number of files or increase the allowed amount by setting the argument max_num_files."
        )
    elif len(filepaths) > 8:
        warnings.warn(
            f"You provided {len(filepaths)} files, which is more than one full orbit (8 files). "
            "Processing might take longer than usual."
        )

    # # Construct filename suffix from orbit/frame numbers
    # orbit_start = str(df["orbit_number"].iloc[0]).zfill(5)
    # orbit_end = str(df["orbit_number"].iloc[-1]).zfill(5)
    # frame_start = df["frame_id"].iloc[0]
    # frame_end = df["frame_id"].iloc[-1]

    # if orbit_start == orbit_end:
    #     oaf_string = (
    #         f"{orbit_start}{frame_start}"
    #         if frame_start == frame_end
    #         else f"{orbit_start}{frame_start}-{frame_end}"
    #     )
    # else:
    #     oaf_string = f"{orbit_start}{frame_start}-{orbit_end}{frame_end}"

    def apply_func(ds: Dataset, i: int) -> Dataset:
        """Apply a processing function to a dataset if specified."""
        if func is None:
            return ds
        if func_inputs is None:
            return func(ds)
        if i < len(func_inputs):
            return func(ds, **func_inputs[i])
        raise IndexError("Too few function inputs provided")

    num_files = len(filepaths)
    ds: xr.Dataset | None = None

    if zoom_at is not None:
        # Zoomed read: select portions of two adjacent frames
        frame_indices = np.unique([int(np.floor(zoom_at)), int(np.ceil(zoom_at))])
        offset = zoom_at - frame_indices[0]
        filepaths = [filepaths[i] for i in frame_indices]

        for i, filepath in enumerate(filepaths):
            with read_product(filepath) as frame_ds:
                frame_ds = apply_func(frame_ds, frame_indices[i])

                # Preserve original dtypes
                original_dtypes = {v: frame_ds[v].dtype for v in frame_ds.variables}

                # Select relevant portion of the frame
                n = len(frame_ds[along_track_dim])
                sel_slice = (
                    slice(int(np.floor(n * offset)), n)
                    if i == 0
                    else slice(0, int(np.ceil(n * offset)))
                )
                frame_ds = frame_ds.sel({along_track_dim: sel_slice})

                # Restore dtypes
                for v, dtype in original_dtypes.items():
                    frame_ds[v] = frame_ds[v].astype(dtype)

                ds = (
                    frame_ds.copy()
                    if ds is None
                    else concat_datasets(
                        ds.copy(), frame_ds.copy(), dim=along_track_dim
                    )
                )

    else:
        # Full read and coarsen each frame
        for i, filepath in enumerate(filepaths):
            with read_product(filepath) as frame_ds:
                frame_ds = apply_func(frame_ds, i)

                if coarsen:
                    original_dtypes = {v: frame_ds[v].dtype for v in frame_ds.variables}

                    coarsen_dims = {along_track_dim: num_files}

                    # Circular mean for longitude
                    lon_coarse = (
                        frame_ds["longitude"]
                        .coarsen(coarsen_dims, boundary="trim")
                        .reduce(circular_mean_np)
                    )

                    # Regular mean for the rest
                    rest = (
                        frame_ds.drop_vars("longitude")
                        .coarsen(coarsen_dims, boundary="trim")
                        .mean()  # type: ignore
                    )

                    # Merge results
                    frame_ds = xr.merge([lon_coarse, rest])

                    for v, dtype in original_dtypes.items():
                        frame_ds[v] = frame_ds[v].astype(dtype)

                ds = (
                    frame_ds
                    if ds is None
                    else concat_datasets(ds, frame_ds, dim=along_track_dim)
                )

    # Set output file sources
    if isinstance(ds, Dataset):
        ds.encoding["sources"] = list(filepaths)
        return ds
    else:
        raise RuntimeError(f"Bad implementation")



import numpy as np
import xarray as xr

from ....constants import (
    DEFAULT_READ_EC_PRODUCT_HEADER,
    DEFAULT_READ_EC_PRODUCT_META,
    DEFAULT_READ_EC_PRODUCT_MODIFY,
)
from ....swath_data.across_track_distance import (
    add_across_track_distance,
    add_nadir_track,
    get_nadir_index,
)
from ....xarray_utils import merge_datasets
from .._rename_dataset_content import rename_common_dims_and_vars, rename_var_info
from ..file_info import FileAgency
from ..header_group import add_header_and_meta_data
from ..science_group import read_science_data


def _get_rgb_from_swir1_nir_vis(ds: xr.Dataset) -> np.ndarray:
    get_min_max = lambda x: np.array([ds[x].quantile(0.01), ds[x].quantile(0.99)])

    r_min, r_max = get_min_max("swir1")
    g_min, g_max = get_min_max("nir")
    b_min, b_max = get_min_max("vis")

    r_w, g_w, b_w = [1.0, 1.0, 1.0]
    r_s, g_s, b_s = [1.0, 1.0, 1.0]

    get_v = lambda x, _w, _s, _min, _max: np.clip(
        _w * (ds[x] - _min) / (_s * (_max - _min)), a_min=0, a_max=1
    ).T

    r_v = get_v("swir1", r_w, r_s, r_min, r_max)
    g_v = get_v("nir", g_w, g_s, g_min, g_max)
    b_v = get_v("vis", b_w, b_s, b_min, b_max)

    rgb = np.stack((r_v, g_v, b_v), axis=2)
    rgb[np.isnan(rgb)] = 0.0

    return rgb


def _add_rgb(ds: xr.Dataset) -> xr.Dataset:
    rgb = _get_rgb_from_swir1_nir_vis(ds)
    # rgb = np.reshape(rgb, (rgb.shape[1], rgb.shape[0], 3))

    ds["rgb"] = (("across_track", "along_track", "rgb_color"), rgb)
    ds["rgb"].attrs["units"] = ""
    ds["rgb"].attrs["long_name"] = "False RGB image"

    return ds


def read_product_mrgr(
    filepath: str,
    modify: bool = DEFAULT_READ_EC_PRODUCT_MODIFY,
    header: bool = DEFAULT_READ_EC_PRODUCT_HEADER,
    meta: bool = DEFAULT_READ_EC_PRODUCT_META,
) -> xr.Dataset:
    """Opens MSI_RGR_1C file as a `xarray.Dataset`."""
    ds = read_science_data(filepath, agency=FileAgency.ESA)

    if not modify:
        return ds

    ds = ds.assign(
        vis=ds["pixel_values"].isel({"band": 0}),
        nir=ds["pixel_values"].isel({"band": 1}),
        swir1=ds["pixel_values"].isel({"band": 2}),
        swir2=ds["pixel_values"].isel({"band": 3}),
        tir1=ds["pixel_values"].isel({"band": 4}),
        tir2=ds["pixel_values"].isel({"band": 5}),
        tir3=ds["pixel_values"].isel({"band": 6}),
        vis_uncertainty=ds["pixel_values_uncertainty"].isel({"band": 0}),
        nir_uncertainty=ds["pixel_values_uncertainty"].isel({"band": 1}),
        swir1_uncertainty=ds["pixel_values_uncertainty"].isel({"band": 2}),
        swir2_uncertainty=ds["pixel_values_uncertainty"].isel({"band": 3}),
        tir1_uncertainty=ds["pixel_values_uncertainty"].isel({"band": 4}),
        tir2_uncertainty=ds["pixel_values_uncertainty"].isel({"band": 5}),
        tir3_uncertainty=ds["pixel_values_uncertainty"].isel({"band": 6}),
        vis_line_quality_status=ds["line_quality_status"].isel({"band": 0}),
        nir_line_quality_status=ds["line_quality_status"].isel({"band": 1}),
        swir1_line_quality_status=ds["line_quality_status"].isel({"band": 2}),
        swir2_line_quality_status=ds["line_quality_status"].isel({"band": 3}),
        tir1_line_quality_status=ds["line_quality_status"].isel({"band": 4}),
        tir2_line_quality_status=ds["line_quality_status"].isel({"band": 5}),
        tir3_line_quality_status=ds["line_quality_status"].isel({"band": 6}),
    )
    ds = rename_var_info(ds, "tir2", "TIR-2", "TIR-2", "")
    ds = ds.drop_vars(
        ["pixel_values", "pixel_values_uncertainty", "line_quality_status"]
    )
    ds = ds.drop_dims("band")

    ds = _add_rgb(ds)

    nadir_idx = get_nadir_index(ds)
    ds = ds.rename({"latitude": "swath_latitude"})
    ds = ds.rename({"longitude": "swath_longitude"})
    ds = add_nadir_track(
        ds,
        nadir_idx,
        swath_lat_var="swath_latitude",
        swath_lon_var="swath_longitude",
        along_track_dim="along_track",
        across_track_dim="across_track",
        nadir_lat_var="latitude",
        nadir_lon_var="longitude",
    )
    ds = add_across_track_distance(
        ds, nadir_idx, swath_lat_var="swath_latitude", swath_lon_var="swath_longitude"
    )

    ds = rename_common_dims_and_vars(
        ds,
        along_track_dim="along_track",
        across_track_dim="across_track",
        track_lat_var="latitude",
        track_lon_var="longitude",
        swath_lat_var="swath_latitude",
        swath_lon_var="swath_longitude",
        time_var="time",
    )

    ds = add_header_and_meta_data(filepath=filepath, ds=ds, header=header, meta=meta)

    return ds



from .curtain import CurtainFigure
from .figure_type import ECKFigure, FigureType
from .line import LineFigure
from .map import MapFigure
from .multi_panel import (
    create_column_subfigures,
    create_fig_layout_map_main_zoom_profile,
)
from .profile import ProfileFigure
from .swath import SwathFigure



import os
from typing import TypeAlias

import numpy as np
import pandas as pd
import xarray as xr
from matplotlib.figure import Figure

from ...utils.config import read_config
from ...utils.ground_sites import get_ground_site
from ...utils.read.product.file_info import ProductDataFrame, get_product_infos
from ...utils.time import TimestampLike, time_to_iso, to_timestamp


def create_filepath(
    filename: str = "",
    filepath: str | None = None,
    ds: xr.Dataset | None = None,
    ds_filepath: str | None = None,
    orbit_and_frame: str | None = None,
    utc_timestamp: TimestampLike | None = None,
    use_utc_creation_timestamp: bool = False,
    site_name: str | None = None,
    hmax: int | float | None = None,
    radius: int | float | None = None,
    extra: str | None = None,
    create_dirs: bool = False,
    resolution: str | None = None,
) -> str:
    if not isinstance(filepath, str):
        config = read_config()
        filepath = os.path.join(config.path_to_images, filename)
    else:
        filepath = os.path.join(filepath, filename)

    df: ProductDataFrame | None = None
    if isinstance(ds, xr.Dataset):
        df = get_product_infos(ds)
    elif isinstance(ds_filepath, str):
        df = get_product_infos(ds_filepath)

    _file_type: str | None = None
    if df is not None:
        orbit_and_frame = df.orbit_and_frame[0]
        utc_timestamp = df.start_sensing_time[0]
        _file_type = df.file_type[0]

    if filepath:
        filename_components = []
        if orbit_and_frame is not None:
            filename_components.append(orbit_and_frame)

        if _file_type is not None:
            filename_components.append(_file_type)

        if utc_timestamp is not None:
            utc_timestamp = time_to_iso(utc_timestamp, format="%Y%m%dT%H%M%SZ")
            filename_components.append(utc_timestamp)

        if use_utc_creation_timestamp == True:
            creation_timestamp = time_to_iso(
                pd.Timestamp.now().utcnow(), format="%Y%m%dT%H%M%SZ"
            )
            filename_components.append(creation_timestamp)

        if site_name is not None:
            try:
                site_name = get_ground_site(site_name).name
            except ValueError as e:
                pass
            filename_components.append(f"site{site_name}")

        if radius is not None:
            radius_string = "rad" + str(int(np.round(radius))) + "m"
            filename_components.append(radius_string)

        if hmax is not None:
            hmax_string = "upto" + str(int(np.round(hmax / 1000))) + "km"
            filename_components.append(hmax_string)

        if resolution is not None:
            if resolution == "" or "high" in resolution.lower():
                filename_components.append("HiRes")
            elif "medium" in resolution.lower():
                filename_components.append("MedRes")
            elif "low" in resolution.lower():
                filename_components.append("LowRes")
            else:
                filename_components.append(resolution)

        if extra is not None:
            filename_components.append(extra)

        basename = os.path.basename(filepath)
        filename_components.append(basename)

        new_basename = "_".join(filename_components)

        dirname = os.path.dirname(filepath)
        if create_dirs and not os.path.exists(dirname):
            os.makedirs(dirname)

        new_filepath = os.path.join(dirname, new_basename)
        new_filepath = os.path.abspath(new_filepath)
        return new_filepath
    else:
        raise ValueError("missing filepath inputs")


def save_plot(
    fig: Figure,
    filename: str = "",
    filepath: str | None = None,
    ds: xr.Dataset | None = None,
    ds_filepath: str | None = None,
    dpi: int | None = None,
    pad_inches: int = 0,
    orbit_and_frame: str | None = None,
    utc_timestamp: TimestampLike | None = None,
    use_utc_creation_timestamp: bool = False,
    site_name: str | None = None,
    hmax: int | float | None = None,
    radius: int | float | None = None,
    extra: str | None = None,
    transparent_outside: bool = False,
    bbox_inches: str = "tight",
    verbose: bool = True,
    print_prefix: str = "",
    create_dirs: bool = False,
    transparent_background: bool = False,
    resolution: str | None = None,
    **kwargs,
):
    if not isinstance(fig, Figure):
        fig = fig.fig

    _stime: str = pd.Timestamp.now().strftime("%Y-%m-%d %H:%M:%S")

    try:
        if transparent_background:
            transparent_outside = True

        new_filepath = create_filepath(
            filename,
            filepath,
            ds,
            ds_filepath,
            orbit_and_frame,
            utc_timestamp,
            use_utc_creation_timestamp,
            site_name,
            hmax,
            radius,
            extra,
            create_dirs,
            resolution,
        )

        if transparent_outside:
            fig.patch.set_alpha(0)
        if transparent_background:
            fig.get_axes()[0].patch.set_alpha(0)
        if verbose:
            print(f"{print_prefix}Saving plot ...", end="\r")
        fig.savefig(
            new_filepath,
            bbox_inches=bbox_inches,
            dpi=dpi,
            pad_inches=pad_inches,
            **kwargs,
        )
        _etime: str = pd.Timestamp.now().strftime("%Y-%m-%d %H:%M:%S")
        _dtime: str = str(pd.Timestamp(_etime) - pd.Timestamp(_stime)).split()[-1]
        if verbose:
            print(f"{print_prefix}Plot saved (time taken {_dtime}): <{new_filepath}>")
    except ValueError as e:
        if verbose:
            print(f"{print_prefix}Did not create plot since no filepath was provided")
